'''
Created on 22-04-2013

@author: Damian Rusinek <damian.rusinek@gmail.com>
'''

from aqopa.model.parser.lex_yacc import LexYaccParserExtension
from aqopa.model import Version, VersionRunProcess, VersionRunHost,\
    MetricsSet


class Builder():
    """
    Builder for creating channel objects
    """
    
    def build_version(self, token):
        """
        version : VERSION IDENTIFIER BLOCKOPEN metrics_sets version_run_hosts BLOCKCLOSE
        """
        v = Version(token[2])
        for metrics_set in token[4]:
            v.metrics_sets.append(metrics_set)
        for run_host in token[5]:
            v.run_hosts.append(run_host)
        return v
    
    def create_metrics_set(self, token):
        """
        metrics_set : SET HOST IDENTIFIER LPARAN IDENTIFIER RPARAN SEMICOLON
                | SET HOST IDENTIFIER LPARAN QUALIFIED_IDENTIFIER RPARAN SEMICOLON
        """
        return MetricsSet(token[3], token[5])
    
    def build_run_host(self, token):
        """
        version_run_host : RUN HOST IDENTIFIER version_channels BLOCKOPEN BLOCKCLOSE
                        | RUN HOST IDENTIFIER version_channels version_repetition BLOCKOPEN BLOCKCLOSE
                        | RUN HOST IDENTIFIER version_channels version_repetition version_repetition_channels BLOCKOPEN BLOCKCLOSE
                        | RUN HOST IDENTIFIER version_channels BLOCKOPEN version_run_processes BLOCKCLOSE
                        | RUN HOST IDENTIFIER version_channels version_repetition BLOCKOPEN version_run_processes BLOCKCLOSE
                        | RUN HOST IDENTIFIER version_channels version_repetition version_repetition_channels BLOCKOPEN version_run_processes BLOCKCLOSE
        """
        run_host = VersionRunHost(token[3])
        if '*' in token[4]:
            run_host.all_channels_active = True
        else:
            run_host.active_channels = token[4]
            
        if len(token) == 8:
            
            if isinstance(token[5], int): # RUN HOST IDENTIFIER version_channels version_repetition BLOCKOPEN BLOCKCLOSE
                run_host.repetitions = token[5]
            else: # RUN HOST IDENTIFIER version_channels BLOCKOPEN version_run_processes BLOCKCLOSE
                run_host.run_processes = token[6]
                
        elif len(token) == 9:
            
            if isinstance(token[6], list): # RUN HOST IDENTIFIER version_channels version_repetition version_repetition_channels BLOCKOPEN BLOCKCLOSE
                run_host.repetitions = token[5]
                run_host.repeated_channels = token[6]
            else: # RUN HOST IDENTIFIER version_channels version_repetition BLOCKOPEN version_run_processes BLOCKCLOSE
                run_host.repetitions = token[5]
                run_host.run_processes = token[7]
            
        elif len(token) == 10:
            run_host.repetitions = token[5]
            run_host.repeated_channels = token[6]
            run_host.run_processes = token[8]
            
            
        return run_host    
    
    def build_run_process(self, token):
        """
        version_run_process : version_run_process_base
                        | version_run_process_base ARROWRIGHT version_run_process_follower
        """
        run_process = token[1]
        if len(token) == 4:
            run_process.follower = token[3]
        return run_process 
    
    def build_run_process_base(self, token):
        """
        version_run_process_base : RUN IDENTIFIER version_subprocesses_list
                                | RUN IDENTIFIER version_subprocesses_list version_repetition
                                | RUN IDENTIFIER version_subprocesses_list version_repetition version_repetition_channels
        """
        run_process = VersionRunProcess(token[2])
        
        if '*' in token[3]:
            run_process.all_subprocesses_active = True
        else:
            for identifier in token[3]:
                run_process.active_subprocesses.append(identifier)
        
        if len(token) > 4:
            run_process.repetitions = token[4]
        
        if len(token) > 5:
            for identifier in token[5]:
                run_process.repeated_channels.append(identifier)
        
        return run_process
    
    def build_run_process_follower(self, token):
        """
        version_run_process_follower : RUN IDENTIFIER version_subprocesses_list
        """
        run_process = VersionRunProcess(token[2])
        
        if '*' in token[3]:
            run_process.all_subprocesses_active = True
        else:
            for identifier in token[3]:
                run_process.active_subprocesses.append(identifier)
        return run_process

class ConfigParserExtension(LexYaccParserExtension):
    """
    Extension for parsing functions
    """
    
    def __init__(self):
        LexYaccParserExtension.__init__(self)

        self.open_blocks_cnt = 0
        
        self.builder = Builder()
        
    ##########################################
    #           RESERVED WORDS
    ##########################################
    
    def word_versions_specification(self, t):
        t.lexer.push_state('versions')
        return t
    
    ##########################################
    #                TOKENS
    ##########################################
    
    def token_block_open(self, t):
        r'{'
        self.open_blocks_cnt += 1
        return t
    
    def token_block_close(self, t):
        r'}'
        self.open_blocks_cnt -= 1
        if self.open_blocks_cnt == 0:
            t.lexer.pop_state()
        return t
    
    def token_sq_lparan(self, t):
        r'\['
        t.lexer.push_state('versionsrepeatedchannels')
        return t
    
    def token_sq_rparan(self, t):
        r'\]'
        t.lexer.pop_state()
        return t
    
    def token_error(self, t):
        self.parser.syntax_errors.append("Line [%s:%s]: Illegal character '%s' \n" % (t.lexer.lineno, t.lexer.lexpos, t.value[0]))
    
    def t_newline(self, t):
        r'\n+'
        t.lexer.lineno += t.value.count("\n")
    
    ##########################################
    #                RULES
    ##########################################
    
    def versions_specification(self, t):
        """
        specification : VERSIONS_SPECIFICATION BLOCKOPEN versions_list BLOCKCLOSE
        """
        pass
    
    
    def versions_list(self, t):
        """
        versions_list : version 
                    | versions_list version
        """
        pass
    
    def version(self, t):
        """
        version : VERSION IDENTIFIER BLOCKOPEN metrics_sets version_run_hosts BLOCKCLOSE
        """
        self.parser.store.versions.append(self.builder.build_version(t))
    
    def metrics_sets(self, t):
        """
        metrics_sets : metrics_set
                    | metrics_sets metrics_set
        """
        if len(t) == 3:
            t[0] = t[1]
            t[0].append(t[2])
        else:
            t[0] = []
            t[0].append(t[1])
    
    def metrics_set(self, t):
        """
        metrics_set : SET HOST IDENTIFIER LPARAN IDENTIFIER RPARAN SEMICOLON
                | SET HOST IDENTIFIER LPARAN QUALIFIED_IDENTIFIER RPARAN SEMICOLON
        """
        t[0] = self.builder.create_metrics_set(t)
    
    def version_run_hosts(self, t):
        """
        version_run_hosts : version_run_host
                        | version_run_hosts version_run_host
        """
        if len(t) == 3:
            t[0] = t[1]
            t[0].append(t[2])
        else:
            t[0] = []
            t[0].append(t[1])
    
    def version_run_host(self, t):
        """
        version_run_host : RUN HOST IDENTIFIER version_channels BLOCKOPEN BLOCKCLOSE
                        | RUN HOST IDENTIFIER version_channels version_repetition BLOCKOPEN BLOCKCLOSE
                        | RUN HOST IDENTIFIER version_channels version_repetition version_repetition_channels BLOCKOPEN BLOCKCLOSE
                        | RUN HOST IDENTIFIER version_channels BLOCKOPEN version_run_processes BLOCKCLOSE
                        | RUN HOST IDENTIFIER version_channels version_repetition BLOCKOPEN version_run_processes BLOCKCLOSE
                        | RUN HOST IDENTIFIER version_channels version_repetition version_repetition_channels BLOCKOPEN version_run_processes BLOCKCLOSE
        """
        t[0] = self.builder.build_run_host(t)
    
    def version_repetition(self, t):
        """
        version_repetition : BLOCKOPEN INTEGER BLOCKCLOSE
        """
        t[0] = t[2]
    
    def version_channels(self, t):
        """
        version_channels : LPARAN RPARAN
                        | LPARAN STAR RPARAN
                        | LPARAN identifiers_list RPARAN
        """
        if len(t) == 3:
            t[0] = []
        elif len(t) == 4:
            if t[2] == '*':
                t[0] = ['*']
            else:
                t[0] = t[2]
    
    def version_repetition_channels(self, t):
        """
        version_repetition_channels : SQLPARAN version_repetition_channels_list SQRPARAN
        """
        t[0] = t[2]
        
    def version_repetition_channels_list(self, t):
        """
        version_repetition_channels_list : version_repetition_channel
                                        | version_repetition_channels_list COMMA version_repetition_channel
        """
        if len(t) == 4:
            t[0] = t[1]
            t[0].append(t[3])
        else:
            t[0] = []
            t[0].append(t[1])
        
    def version_repetition_channel(self, t):
        """
        version_repetition_channel : QUALIFIED_IDENTIFIER 
                                | IDENTIFIER
        """
        t[0] = t[1]
    
    def version_run_processes(self, t):
        """
        version_run_processes : version_run_process
                            | version_run_processes version_run_process
        """
        if len(t) == 3:
            t[0] = t[1]
            t[0].append(t[2])
        else:
            t[0] = []
            t[0].append(t[1])
    
    def version_run_process(self, t):
        """
        version_run_process : version_run_process_base
                        | version_run_process_base ARROWRIGHT version_run_process_follower
        """
        t[0] = self.builder.build_run_process(t)
    
    def version_run_process_base(self, t):
        """
        version_run_process_base : RUN IDENTIFIER version_subprocesses_list
                                | RUN IDENTIFIER version_subprocesses_list version_repetition
                                | RUN IDENTIFIER version_subprocesses_list version_repetition version_repetition_channels
        """
        t[0] = self.builder.build_run_process_base(t)
    
    def version_run_process_follower(self, t):
        """
        version_run_process_follower : RUN IDENTIFIER version_subprocesses_list
        """
        t[0] = self.builder.build_run_process_follower(t)
    
    def version_subprocesses_list(self, t):
        """
        version_subprocesses_list : LPARAN RPARAN
                        | LPARAN STAR RPARAN
                        | LPARAN identifiers_list RPARAN
        """
        if len(t) == 3:
            t[0] = []
        elif len(t) == 4:
            if t[2] == '*':
                t[0] = ['*']
            else:
                t[0] = t[2]
    
    def _extend(self):
        
        self.parser.add_state('versions', 'inclusive')
        self.parser.add_state('versionsrepeatedchannels', 'exclusive')
        
        self.parser.add_reserved_word('versions', 'VERSIONS_SPECIFICATION', func=self.word_versions_specification)
        self.parser.add_reserved_word('version', 'VERSION', state='versions')
        self.parser.add_reserved_word('run', 'RUN', state='versions')
        self.parser.add_reserved_word('set', 'SET', state='versions')
        self.parser.add_reserved_word('host', 'HOST', state='versions')

        self.parser.add_token('BLOCKOPEN', func=self.token_block_open, states=['versions'])
        self.parser.add_token('BLOCKCLOSE', func=self.token_block_close, states=['versions'])
        self.parser.add_token('ARROWRIGHT', r'\-\>', states=['versions'])
        
        self.parser.add_token('SQLPARAN', func=self.token_sq_lparan, states=['versions'])
        self.parser.add_token('IDENTIFIER', r'[_a-zA-Z][_a-zA-Z0-9]*', states=['versionsrepeatedchannels'])
        self.parser.add_token('QUALIFIED_IDENTIFIER', r'[_a-zA-Z][_a-zA-Z0-9]*(\.[0-9][0-9]*)+', 
                              states=['versionsrepeatedchannels'])
        self.parser.add_token('COMMA', r',', states=['versionsrepeatedchannels'])
        self.parser.add_token('SQRPARAN', func=self.token_sq_rparan, states=['versionsrepeatedchannels'])
        
        self.parser.add_token('error', func=self.token_error, states=['versionsrepeatedchannels'], include_in_tokens=False)
        self.parser.add_token('ignore', "\t ", states=['versionsrepeatedchannels'], include_in_tokens=False)
        self.parser.add_token('newline', func=self.t_newline,  states=['versionsrepeatedchannels'], include_in_tokens=False)

        self.parser.add_rule(self.versions_specification)
        self.parser.add_rule(self.versions_list)
        self.parser.add_rule(self.version)
        self.parser.add_rule(self.metrics_sets)
        self.parser.add_rule(self.metrics_set)
        self.parser.add_rule(self.version_run_hosts)
        self.parser.add_rule(self.version_run_host)
        self.parser.add_rule(self.version_repetition)
        self.parser.add_rule(self.version_channels)
        self.parser.add_rule(self.version_repetition_channels)
        self.parser.add_rule(self.version_repetition_channels_list)
        self.parser.add_rule(self.version_repetition_channel)
        self.parser.add_rule(self.version_run_processes)
        self.parser.add_rule(self.version_run_process)
        self.parser.add_rule(self.version_run_process_base)
        self.parser.add_rule(self.version_run_process_follower)
        self.parser.add_rule(self.version_subprocesses_list)
        

    


'''
Created on 22-04-2013

@author: Damian Rusinek <damian.rusinek@gmail.com>
'''
from aqopa.model.parser.lex_yacc import LexYaccParserExtension

class ModelParserExtension(LexYaccParserExtension):
    
    def __init__(self):
        LexYaccParserExtension.__init__(self)
        
    ##########################################
    #           RESERVED WORDS
    ##########################################
    
    def word_bool(self, t):
        t.value = bool(t.value)
        return t
    
    ##########################################
    #                TOKENS
    ##########################################
    
    def token_identifier(self, t):
        r'[_a-zA-Z][_a-zA-Z0-9]*'
        
        words = self.parser.get_reserved_words()
        
        states_stack = []
        states_stack.extend(t.lexer.lexstatestack)
        states_stack.append(t.lexer.current_state())
        
        i = len(states_stack)-1
        while i >= 0:
            state = states_stack[i]
            if state in words:
                state_words = words[state]
                for state_word in state_words:
                    
                    tvalue = t.value
                    state_word_value = state_word
                    word_tuple = state_words[state_word]
                    
                    # if not case sensitive
                    if not word_tuple[2]:
                        tvalue = tvalue.lower()
                        state_word_value = state_word_value.lower()
                    
                    if tvalue == state_word_value:
                        # If function exists
                        if word_tuple[1]:
                            t = word_tuple[1](t)
                        t.type = word_tuple[0]
                        break
            i -= 1
            
        return t
    
    def token_integer(self, t):
        r'0|[1-9][0-9]*'
        
        t.value = int(t.value)
        return t
    
    def token_float(self, t):
        r'[1-9][0-9]*(\.[0-9]+)?'
        
        t.value = float(t.value)
        return t
        
    
    ##########################################
    #                RULES
    ##########################################
    
    def rule_model(self, t):
        """
        model : specification 
            | model specification
        """
        pass

    def rule_identifiers_list(self, t):
        """
        identifiers_list : IDENTIFIER
                    | identifiers_list COMMA IDENTIFIER
        """
        if len(t) == 2:
            t[0] = [t[1]]
        else:
            t[0] = t[1]
            t[0].append(t[3])

    def rule_qualified_identifiers_list(self, t):
        """
        qualified_identifiers_list : QUALIFIED_IDENTIFIER
                    | qualified_identifiers_list COMMA QUALIFIED_IDENTIFIER
        """
        if len(t) == 2:
            t[0] = [t[1]]
        else:
            t[0] = t[1]
            t[0].append(t[3])
    
    def _extend(self):
        
        self.parser.add_reserved_word('true', 'BOOL', func=self.word_bool)
        self.parser.add_reserved_word('false', 'BOOL', func=self.word_bool)
        
        self.parser.add_token('COMMA', r',')
        self.parser.add_token('INTEGER', func=self.token_integer)
        self.parser.add_token('QUALIFIED_IDENTIFIER', r'[_a-zA-Z][_a-zA-Z0-9]*(\.[1-9][0-9]*)+')
        self.parser.add_token('IDENTIFIER', func=self.token_identifier)
        self.parser.add_token('TEXT', r'[-_A-Za-z0-9 ]+')  
        
        self.parser.add_token('SEMICOLON', r';')
        self.parser.add_token('COLON', r':',)
        self.parser.add_token('STAR', r'\*')
        self.parser.add_token('EQUAL', r'\=')
        
        self.parser.add_token('LPARAN', r'\(')
        self.parser.add_token('RPARAN', r'\)')
        self.parser.add_token('SQLPARAN', r'\[')
        self.parser.add_token('SQRPARAN', r'\]')
        self.parser.add_token('BLOCKOPEN', r'{')
        self.parser.add_token('BLOCKCLOSE', r'}')
        
        self.parser.add_rule(self.rule_model)
        self.parser.add_rule(self.rule_identifiers_list)
        self.parser.add_rule(self.rule_qualified_identifiers_list)
        
        self.parser.start_symbol = 'model'

class MetricsParserExtension(LexYaccParserExtension):
    """
    """
    
    def __init__(self):
        LexYaccParserExtension.__init__(self)
        
    ##########################################
    #           RESERVED WORDS
    ##########################################
    
    def word_bool(self, t):
        t.value = bool(t.value)
        return t
    
    ##########################################
    #                TOKENS
    ##########################################
    
    def token_identifier(self, t):
        r'[_a-zA-Z][_a-zA-Z0-9]*'
        
        words = self.parser.get_reserved_words()
        
        states_stack = []
        states_stack.extend(t.lexer.lexstatestack)
        states_stack.append(t.lexer.current_state())
        
        i = len(states_stack)-1
        while i >= 0:
            state = states_stack[i]
            if state in words:
                state_words = words[state]
                for state_word in state_words:
                    
                    tvalue = t.value
                    state_word_value = state_word
                    word_tuple = state_words[state_word]
                    
                    # if not case sensitive
                    if not word_tuple[2]:
                        tvalue = tvalue.lower()
                        state_word_value = state_word_value.lower()
                    
                    if tvalue == state_word_value:
                        # If function exists
                        if word_tuple[1]:
                            t = word_tuple[1](t)
                        t.type = word_tuple[0]
                        break
            i -= 1
            
        return t
    
    def token_integer(self, t):
        r'0|[1-9][0-9]*'
        
        t.value = int(t.value)
        return t
    
    def token_float(self, t):
        r'[1-9][0-9]*(\.[0-9]+)?'
        
        t.value = float(t.value)
        return t
        
    
    ##########################################
    #                RULES
    ##########################################
    
    def rule_metrics(self, t):
        """
        metrics : specification 
            | metrics specification
        """
        pass

    def rule_identifiers_list(self, t):
        """
        identifiers_list : IDENTIFIER
                    | identifiers_list COMMA IDENTIFIER
        """
        if len(t) == 2:
            t[0] = [t[1]]
        else:
            t[0] = t[1]
            t[0].append(t[3])

    def rule_qualified_identifiers_list(self, t):
        """
        qualified_identifiers_list : QUALIFIED_IDENTIFIER
                    | qualified_identifiers_list COMMA QUALIFIED_IDENTIFIER
        """
        if len(t) == 2:
            t[0] = [t[1]]
        else:
            t[0] = t[1]
            t[0].append(t[3])
    
    def _extend(self):
        
        self.parser.add_token('COMMA', r',')
        self.parser.add_token('INTEGER', func=self.token_integer)
        self.parser.add_token('QUALIFIED_IDENTIFIER', r'[_a-zA-Z][_a-zA-Z0-9]*(\.[1-9][0-9]*)+')
        self.parser.add_token('IDENTIFIER', func=self.token_identifier)
        self.parser.add_token('TEXT', r'[-_A-Za-z0-9 ]+')  
        
        self.parser.add_token('SEMICOLON', r';')
        self.parser.add_token('COLON', r':',)
        self.parser.add_token('STAR', r'\*')
        self.parser.add_token('EQUAL', r'\=')
        
        self.parser.add_token('LPARAN', r'\(')
        self.parser.add_token('RPARAN', r'\)')
        self.parser.add_token('SQLPARAN', r'\[')
        self.parser.add_token('SQRPARAN', r'\]')
        self.parser.add_token('BLOCKOPEN', r'{')
        self.parser.add_token('BLOCKCLOSE', r'}')
        
        self.parser.add_rule(self.rule_metrics)
        self.parser.add_rule(self.rule_identifiers_list)
        self.parser.add_rule(self.rule_qualified_identifiers_list)
        
        self.parser.start_symbol = 'metrics'


class ConfigParserExtension(LexYaccParserExtension):
    """
    """
    
    def __init__(self):
        LexYaccParserExtension.__init__(self)
        
    ##########################################
    #           RESERVED WORDS
    ##########################################
    
    def word_bool(self, t):
        t.value = bool(t.value)
        return t
    
    ##########################################
    #                TOKENS
    ##########################################
    
    def token_identifier(self, t):
        r'[_a-zA-Z][_a-zA-Z0-9]*'
        
        words = self.parser.get_reserved_words()
        
        states_stack = []
        states_stack.extend(t.lexer.lexstatestack)
        states_stack.append(t.lexer.current_state())
        
        i = len(states_stack)-1
        while i >= 0:
            state = states_stack[i]
            if state in words:
                state_words = words[state]
                for state_word in state_words:
                    
                    tvalue = t.value
                    state_word_value = state_word
                    word_tuple = state_words[state_word]
                    
                    # if not case sensitive
                    if not word_tuple[2]:
                        tvalue = tvalue.lower()
                        state_word_value = state_word_value.lower()
                    
                    if tvalue == state_word_value:
                        # If function exists
                        if word_tuple[1]:
                            t = word_tuple[1](t)
                        t.type = word_tuple[0]
                        break
            i -= 1
            
        return t
    
    def token_integer(self, t):
        r'0|[1-9][0-9]*'
        
        t.value = int(t.value)
        return t
    
    def token_float(self, t):
        r'[1-9][0-9]*(\.[0-9]+)?'
        
        t.value = float(t.value)
        return t
        
    
    ##########################################
    #                RULES
    ##########################################
    
    def rule_configuration(self, t):
        """
        configuration : specification 
            | configuration specification
        """
        pass

    def rule_identifiers_list(self, t):
        """
        identifiers_list : IDENTIFIER
                    | identifiers_list COMMA IDENTIFIER
        """
        if len(t) == 2:
            t[0] = [t[1]]
        else:
            t[0] = t[1]
            t[0].append(t[3])

    def rule_qualified_identifiers_list(self, t):
        """
        qualified_identifiers_list : QUALIFIED_IDENTIFIER
                    | qualified_identifiers_list COMMA QUALIFIED_IDENTIFIER
        """
        if len(t) == 2:
            t[0] = [t[1]]
        else:
            t[0] = t[1]
            t[0].append(t[3])
    
    def _extend(self):
        
        self.parser.add_reserved_word('true', 'BOOL', func=self.word_bool)
        self.parser.add_reserved_word('false', 'BOOL', func=self.word_bool)
        
        self.parser.add_token('COMMA', r',')
        self.parser.add_token('INTEGER', func=self.token_integer)
        self.parser.add_token('QUALIFIED_IDENTIFIER', r'[_a-zA-Z][_a-zA-Z0-9]*(\.[1-9][0-9]*)+')
        self.parser.add_token('IDENTIFIER', func=self.token_identifier)
        self.parser.add_token('TEXT', r'[-_A-Za-z0-9 ]+')  
        
        self.parser.add_token('SEMICOLON', r';')
        self.parser.add_token('COLON', r':',)
        self.parser.add_token('STAR', r'\*')
        self.parser.add_token('EQUAL', r'\=')
        
        self.parser.add_token('LPARAN', r'\(')
        self.parser.add_token('RPARAN', r'\)')
        self.parser.add_token('SQLPARAN', r'\[')
        self.parser.add_token('SQRPARAN', r'\]')
        self.parser.add_token('BLOCKOPEN', r'{')
        self.parser.add_token('BLOCKCLOSE', r'}')
        
        self.parser.add_rule(self.rule_configuration)
        self.parser.add_rule(self.rule_identifiers_list)
        self.parser.add_rule(self.rule_qualified_identifiers_list)
        
        self.parser.start_symbol = 'configuration'




'''
Created on 22-04-2013

@author: Damian Rusinek <damian.rusinek@gmail.com>
'''

from aqopa.model.parser.lex_yacc import LexYaccParserExtension
from aqopa.model import BooleanExpression, TupleExpression,\
    TupleElementExpression, IdentifierExpression, ComparisonExpression,\
    CallFunctionExpression


class Builder():

    def build_expression_call_function(self, token):
        """
        expression_call_function : IDENTIFIER LPARAN RPARAN 
                                | IDENTIFIER LPARAN expression_function_arguments RPARAN
                                | IDENTIFIER LPARAN RPARAN SQLPARAN expression_function_qop_arguments SQRPARAN
                                | IDENTIFIER LPARAN expression_function_arguments RPARAN SQLPARAN expression_function_qop_arguments SQRPARAN
        """
        arguments = []
        qop_arguments = []
        
        if len(token) == 5:
            arguments = token[3]
        elif len(token) == 7:
            qop_arguments = token[5]
        elif len(token) == 8:
            arguments = token[3]
            qop_arguments = token[6]
            
        return CallFunctionExpression(token[1], arguments, qop_arguments)
        

class ModelParserExtension(LexYaccParserExtension):
    """
    Extension for parsing functions
    """
    
    def __init__(self):
        LexYaccParserExtension.__init__(self)
        
        self.builder = Builder()
        
    ##########################################
    #           RESERVED WORDS
    ##########################################
    
    ##########################################
    #                TOKENS
    ##########################################

    ##########################################
    #                RULES
    ##########################################
    
    def expression_conditional(self, t):
        """
        expression_conditional : expression_comaprison
                        | BOOL
        """
        if isinstance(t[1], bool):
            t[0] = BooleanExpression(t[1])
        else:
            t[0] = t[1]
    
    def expression_tuple(self, t):
        """
        expression_tuple : LPARAN RPARAN 
                        | LPARAN expression_function_arguments RPARAN
        """
        elements = []
        if len(t) > 3:
            elements = t[2]
        t[0] = TupleExpression(elements)
    
    def expression_tuple_element(self, t):
        """
        expression_tuple_element : IDENTIFIER SQLPARAN INTEGER SQRPARAN
        """
        t[0] = TupleElementExpression(t[1], t[3])
    
    def expression_function_arguments(self, t):
        """
        expression_function_arguments : expression_simple
                                    | expression_function_arguments COMMA expression_simple
        """
        if len(t) == 2:
            t[0] = []
            t[0].append(t[1])
        else:
            t[0] = t[1]
            t[0].append(t[3])
        
    def expression_call_function(self, t):
        """
        expression_call_function : IDENTIFIER LPARAN RPARAN 
                                | IDENTIFIER LPARAN expression_function_arguments RPARAN
                                | IDENTIFIER LPARAN RPARAN SQLPARAN expression_function_qop_arguments SQRPARAN
                                | IDENTIFIER LPARAN expression_function_arguments RPARAN SQLPARAN expression_function_qop_arguments SQRPARAN
        """
        t[0] = self.builder.build_expression_call_function(t)
    
    def expression_function_qop_arguments(self, t):
        """
        expression_function_qop_arguments : TEXT 
                                        | expression_function_qop_arguments COMMA TEXT 
        """
        if len(t) == 2:
            t[0] = []
            t[0].append(t[1])
        else:
            t[0] = t[1]
            t[0].append(t[3])
    
    def expression_simple(self, t):
        """
        expression_simple : expression_call_function
                        | expression_tuple_element
                        | expression_tuple
                        | IDENTIFIER
                        | BOOL
        """
        if isinstance(t[1], bool):
            t[0] = BooleanExpression(t[1])
        elif isinstance(t[1], str) or isinstance(t[1], unicode):
            t[0] = IdentifierExpression(t[1])
        else:
            t[0] = t[1]
    
    def expression_comaprison(self, t):
        """
        expression_comaprison : expression_simple EQUAL EQUAL expression_simple
        """
        t[0] = ComparisonExpression(t[1], t[4])
    
    def _extend(self):
        
        self.parser.add_rule(self.expression_conditional)
        self.parser.add_rule(self.expression_tuple)
        self.parser.add_rule(self.expression_tuple_element)
        self.parser.add_rule(self.expression_function_arguments)
        self.parser.add_rule(self.expression_call_function)
        self.parser.add_rule(self.expression_function_qop_arguments)
        self.parser.add_rule(self.expression_simple)
        self.parser.add_rule(self.expression_comaprison)
        

    


#!/usr/bin/env python
'''
Created on 22-04-2013

@author: Damian Rusinek <damian.rusinek@gmail.com>
'''
import os
import sys
import threading
import time

from aqopa.model.parser import ModelParserException,\
    MetricsParserException, ConfigurationParserException
from aqopa.app import Builder, ConsoleInterpreter
from aqopa.simulator import EnvironmentDefinitionException
from aqopa.module import timeanalysis

class ProgressThread(threading.Thread):

    def __init__(self, f, interpreter, *args, **kwargs):
        super(ProgressThread, self).__init__(*args, **kwargs)
        self.file = f
        self.interpreter = interpreter
        self.signs = '|/-\\'
        self.sign_no = 0
        
    def get_progress(self):
        all_progress = 0.0
        sum_progress = 0.0
        for thr in self.interpreter.threads:
            all_progress += 1
            sum_progress += thr.simulator.context.get_progress()
        progress = 0
        if all_progress > 0:
            progress = sum_progress / all_progress
        return progress
            
    def run(self):
        
        progress = self.get_progress()
        while progress < 1:
            self.print_progressbar(progress)
            time.sleep(0.1)
            progress = self.get_progress()
        self.print_progressbar(progress)
        
        
    def print_progressbar(self, progress):
            """
            Prints the formatted progressbar showing the progress of simulation. 
            """
            self.sign_no += 1
            sign = self.signs[self.sign_no % len(self.signs)]
            
            percentage = str(int(round(progress*100))) + '%'
            percentage = (' ' * (5-len(percentage))) + percentage
            
            bar = ('#' * int(round(progress*20))) + (' ' * (20 - int(round(progress*20))))
            
            self.file.write("\r%c[%s] %s" % (sign, bar, percentage))
            self.file.flush()
        

def run(qopml_model, qopml_metrics, qopml_configuration, 
         save_states = False, debug = False, show_progressbar = False):

    ############### DEBUG ###############    
    if debug:
        builder = Builder()
        store = builder.build_store()
        parser = builder.build_model_parser(store, [])
        parser.lexer.input(qopml_model)
        while True:
            print  parser.lexer.current_state()
            tok = parser.lexer.token()
            if not tok:
                break
            print tok
            print ""
        print 'Errors: ' + str(parser.get_syntax_errors())
        print ""
        print ""
    
    #####################################
    interpreter = ConsoleInterpreter()
    try:
        interpreter.set_qopml_model(qopml_model)
        interpreter.set_qopml_metrics(qopml_metrics)
        interpreter.set_qopml_config(qopml_configuration)
        
        interpreter.register_qopml_module(timeanalysis.Module())
        
        interpreter.parse()
        interpreter.prepare()
        
        if save_states:
            for simulator in interpreter.simulators: 
                interpreter.save_states_to_file(simulator)
        
        if show_progressbar:
            progressbar_thread = ProgressThread(sys.stdout, interpreter)
            progressbar_thread.start()
        
        interpreter.run()

    except EnvironmentDefinitionException, e:
        sys.stderr.write('Error on creating environment: %s\n' % e)
        if len(e.errors) > 0:
            sys.stderr.write('Errors:\n')
            sys.stderr.write('\n'.join(e.errors))
            sys.stderr.write('\n')
    except ModelParserException, e:
        sys.stderr.write('Model parsing error: %s\n' % e)
        if len(e.syntax_errors):
            sys.stderr.write('Syntax errors:\n')
            sys.stderr.write('\n'.join(e.syntax_errors))
            sys.stderr.write('\n')
    except MetricsParserException, e:
        sys.stderr.write('Metrics parsing error: %s\n' % e)
        if len(e.syntax_errors):
            sys.stderr.write('Syntax errors:\n')
            sys.stderr.write('\n'.join(e.syntax_errors))
            sys.stderr.write('\n')
    except ConfigurationParserException, e:
        sys.stderr.write('Configuration parsing error: %s\n' % e)
        if len(e.syntax_errors):
            sys.stderr.write('Syntax errors:\n')
            sys.stderr.write('\n'.join(e.syntax_errors))
            sys.stderr.write('\n')
        



'''
Created on 22-04-2013

@author: Damian Rusinek <damian.rusinek@gmail.com>
'''

from aqopa.model.parser.lex_yacc import LexYaccParserExtension
from aqopa.model import MetricsConfiguration, MetricsData,\
    MetricsPrimitiveBlock, MetricsPrimitiveHeader, MetricsPrimitive, \
    MetricsServiceParam


class Builder():

    def create_metrics_configuration(self, token):
        """
        metrics_configuration : METRICS_CONFIGURATION LPARAN IDENTIFIER RPARAN BLOCKOPEN metrics_configuration_params BLOCKCLOSE
        """
        return MetricsConfiguration(token[3], token[6])
    
    def create_metrics_data(self, token):
        """
        metrics_data : METRICS_DATA LPARAN IDENTIFIER RPARAN BLOCKOPEN metrics_data_blocks BLOCKCLOSE
                    | METRICS_DATA PLUS LPARAN QUALIFIED_IDENTIFIER RPARAN BLOCKOPEN metrics_data_blocks BLOCKCLOSE
                    | METRICS_DATA STAR LPARAN QUALIFIED_IDENTIFIER RPARAN BLOCKOPEN metrics_data_blocks BLOCKCLOSE
        """
        
        if len(token) == 8:
            name = token[3]
            blocks = token[6]
            md = MetricsData(name, blocks)
        elif len(token) == 9:
            name = token[4]
            blocks = token[7]
            plus = token[2] == '+' 
            star = token[2] == '*'
            md = MetricsData(name, blocks, plus, star)
            
        return md 
    
    def create_metrics_block(self, token):
        """
        metrics_data_block : metrics_primitives_head metrics_primitives 
        """
        return MetricsPrimitiveBlock(token[1], token[2])
    
    def create_metrics_header(self, token):
        """
        metrics_primitives_head : METRICS_PRIMITIVES_HEAD metrics_params metrics_services_params SEMICOLON
        """
        return MetricsPrimitiveHeader(token[2], token[3])
    
    def create_metrics_primitive(self, token):
        """
        metrics_primitive : METRICS_PRIMITIVE metrics_primitive_arguments SEMICOLON
        """
        return MetricsPrimitive(token[2])
    
    def create_metrics_services_param_size(self, token):
        """
        metrics_services_param : SQLPARAN SIZE COLON metrics_services_size_type_unit LPARAN metrics_services_size_unit RPARAN SQRPARAN
                            | SQLPARAN SIZE COLON metrics_services_size_type_non_unit SQRPARAN
        """
        unit = token[6] if len(token) == 9 else None
        return MetricsServiceParam(token[2], token[4], unit)
        

class MetricsParserExtension(LexYaccParserExtension):
    """
    Extension for parsing metrics
    """
    
    def __init__(self):
        LexYaccParserExtension.__init__(self)
        
        self.builder = Builder()
        
    ##########################################
    #           RESERVED WORDS
    ##########################################
    
    def word_metrics_specification(self, t):
        t.lexer.push_state('metrics')
        return t
    
    def word_metrics_data(self, t):
        t.lexer.push_state('metricsdata')
        return t
    
    def word_metrics_configuration(self, t):
        t.lexer.push_state('metricsconfiguration')
        return t
    
    def word_metricsdata_primhead(self, t):
        t.lexer.push_state('metricsprimhead')
        return t
    
    def word_metricsdata_primitive(self, t):
        t.lexer.push_state('metricsprimitive')
        return t
    
    ##########################################
    #                TOKENS
    ##########################################
    
    def token_metrics_block_close(self, t):
        r"\}"
        t.lexer.pop_state()
        return t

    def token_metricsconfiguration_block_open(self, t):
        r"\{"
        t.lexer.push_state('metricsconfigurationparams')
        return t

    def token_metricsconfigurationparams_block_close(self, t):
        r"\}"
        t.lexer.pop_state()
        t.lexer.pop_state()
        return t
    
    def token_metricsconfigurationparams_equal(self, t):
        r'\='
        t.lexer.push_state('metricsconfigurationparamsvalue')
        return t
    
    def token_metricsconfigurationparamsvalue_paramvalue(self, t):
        r'[^;]+'
        t.lexer.pop_state()
        return t
    
    def token_metricsdata_block_close(self, t):
        r"\}"
        t.lexer.pop_state()
        return t
    
    def token_metricsprimhead_semicolon(self, t):
        r"\;"
        t.lexer.pop_state()
        return t
    
    def token_metricsprimitive_sqlparan(self, t):
        r"\["
        t.lexer.push_state('metricsprimitiveargumentvalue')
        return t
    
    def token_metricsprimitive_semicolon(self, t):
        r"\;"
        t.lexer.pop_state()
        return t
    
    def token_metricsprimitiveargumentvalue_argumentvalue(self, t):
        r'[^]]+'
        t.lexer.pop_state()
        return t
    
    def token_error(self, t):
        self.parser.syntax_errors.append("Line [%s:%s]: Illegal character '%s' \n" % (t.lexer.lineno, t.lexer.lexpos, t.value[0]))
    
    def t_newline(self, t):
        r'\n+'
        t.lexer.lineno += t.value.count("\n")
    

    ##########################################
    #                RULES
    ##########################################
    
    
    def metrics_specification(self, t):
        """
        specification : METRICS_SPECIFICATION BLOCKOPEN metrics_configurations metrics_datas BLOCKCLOSE
        """
        pass
    
    def metrics_configurations(self, t):
        """
        metrics_configurations : metrics_configuration
                            | metrics_configurations metrics_configuration
        """
        pass
    
    def metrics_configuration(self, t):
        """
        metrics_configuration : METRICS_CONFIGURATION LPARAN IDENTIFIER RPARAN BLOCKOPEN metrics_configuration_params BLOCKCLOSE
        """
        self.parser.store.metrics_configurations.append(self.builder.create_metrics_configuration(t))
    
    def metrics_configuration_params(self, t):
        """
        metrics_configuration_params : metrics_configuration_param
                                    | metrics_configuration_params metrics_configuration_param
        """
        if len(t) == 2:
            t[0] = []
            t[0].append(t[1])
        else:
            t[0] = t[1]
            t[0].append(t[2])
    
    def metrics_configuration_param(self, t):
        """
        metrics_configuration_param : IDENTIFIER EQUAL PARAMVALUE SEMICOLON
        """
        t[0] = (t[1], t[2])
    
    def metrics_datas(self, t):
        """
        metrics_datas : metrics_data 
                    | metrics_datas metrics_data
        """
        pass
        
    def metrics_data(self, t):
        """
        metrics_data : METRICS_DATA LPARAN IDENTIFIER RPARAN BLOCKOPEN metrics_data_blocks BLOCKCLOSE
                    | METRICS_DATA PLUS LPARAN QUALIFIED_IDENTIFIER RPARAN BLOCKOPEN metrics_data_blocks BLOCKCLOSE
                    | METRICS_DATA STAR LPARAN QUALIFIED_IDENTIFIER RPARAN BLOCKOPEN metrics_data_blocks BLOCKCLOSE
        """
        self.parser.store.metrics_datas.append(self.builder.create_metrics_data(t))
        
    def metrics_data_blocks(self, t):
        """
        metrics_data_blocks : metrics_data_block
                            | metrics_data_blocks HASH metrics_data_block
        """
        if len(t) == 2:
            t[0] = []
            t[0].append(t[1])
        else:
            t[0] = t[1]
            t[0].append(t[3])
    
    def metrics_data_block(self, t):
        """
        metrics_data_block : metrics_primitives_head metrics_primitives 
        """
        t[0] = self.builder.create_metrics_block(t)
    
    def metrics_primitives_head(self, t):
        """
        metrics_primitives_head : METRICS_PRIMITIVES_HEAD metrics_params metrics_services_params SEMICOLON
        """
        t[0] = self.builder.create_metrics_header(t)
    
    def metrics_params(self, t):
        """
        metrics_params : metrics_param
                        | metrics_params metrics_param
        """
        if len(t) == 2:
            t[0] = []
            t[0].append(t[1])
        else:
            t[0] = t[1]
            t[0].append(t[2])
    
    def metrics_param(self, t):
        """
        metrics_param : SQLPARAN IDENTIFIER SQRPARAN
        """
        t[0] = t[2]
    
    def metrics_services_params(self, t):
        """
        metrics_services_params : metrics_services_param
                            | metrics_services_params metrics_services_param
        """
        if len(t) == 2:
            t[0] = []
            t[0].append(t[1])
        else:
            t[0] = t[1]
            t[0].append(t[2])

    def metrics_primitives(self, t):
        """
        metrics_primitives : metrics_primitive
                            | metrics_primitives metrics_primitive
        """
        if len(t) == 2:
            t[0] = []
            t[0].append(t[1])
        else:
            t[0] = t[1]
            t[0].append(t[2])
    
    def metrics_primitive(self, t):
        """
        metrics_primitive : METRICS_PRIMITIVE metrics_primitive_arguments SEMICOLON
        """
        t[0] = self.builder.create_metrics_primitive(t)
    
    def metrics_primitive_arguments(self, t):
        """
        metrics_primitive_arguments : metrics_primitive_argument
                                | metrics_primitive_arguments metrics_primitive_argument
        """
        if len(t) == 2:
            t[0] = []
            t[0].append(t[1])
        else:
            t[0] = t[1]
            t[0].append(t[2])
        
    def metrics_primitive_argument(self, t):
        """
        metrics_primitive_argument : SQLPARAN ARGUMENTVALUE SQRPARAN
        """
        t[0] = t[2]

    #######################
    #    Metrics Size
    #######################
    
    def metrics_services_param_size(self, t):
        """
        metrics_services_param : SQLPARAN SIZE COLON metrics_services_size_type_unit LPARAN metrics_services_size_unit RPARAN SQRPARAN
                            | SQLPARAN SIZE COLON metrics_services_size_type_non_unit SQRPARAN
        """
        t[0] = self.builder.create_metrics_services_param_size(t)
    
    def metrics_services_size_type_unit(self, t):
        """
        metrics_services_size_type_unit : EXACT
        """
        t[0] = t[1].lower()
    
    def metrics_services_size_type_non_unit(self, t):
        """
        metrics_services_size_type_non_unit : RATIO
        """
        t[0] = t[1].lower()
    
    def metrics_services_size_unit(self, t):
        """
        metrics_services_size_unit : B
        """
        t[0] = t[1].lower()
    
    
    def _extend(self):
        
        self.parser.add_state('metrics', 'inclusive')
        self.parser.add_state('metricsdata', 'inclusive')
        self.parser.add_state('metricsprimhead', 'inclusive')
        
        self.parser.add_state('metricsconfiguration', 'inclusive')
        self.parser.add_state('metricsconfigurationparams', 'exclusive')
        self.parser.add_state('metricsconfigurationparamsvalue', 'exclusive')
        
        self.parser.add_state('metricsprimitive', 'inclusive')
        self.parser.add_state('metricsprimitiveargumentvalue', 'exclusive')

        self.parser.add_reserved_word('metrics', 'METRICS_SPECIFICATION', func=self.word_metrics_specification,)
        self.parser.add_reserved_word('conf', 'METRICS_CONFIGURATION', func=self.word_metrics_configuration, state='metrics')
        self.parser.add_reserved_word('data', 'METRICS_DATA', func=self.word_metrics_data, state='metrics')
        self.parser.add_reserved_word('primhead', 'METRICS_PRIMITIVES_HEAD', func=self.word_metricsdata_primhead, state='metricsdata')
        self.parser.add_reserved_word('primitive', 'METRICS_PRIMITIVE', func=self.word_metricsdata_primitive, state='metricsdata')
    
        self.parser.add_token('QUALIFIED_IDENTIFIER', r'[_a-zA-Z][_a-zA-Z0-9]*(\.[1-9][0-9]*)+', states=['metrics', 'metricsdata'])
        self.parser.add_token('BLOCKCLOSE', func=self.token_metrics_block_close, states=['metrics'])

        # METRICS CONFIGURATION    
        self.parser.add_token('BLOCKOPEN', func=self.token_metricsconfiguration_block_open, states=['metricsconfiguration'])
        
        # Metrics configuration params state
        self.parser.add_token('error', func=self.token_error, states=['metricsconfigurationparams'], include_in_tokens=False)
        self.parser.add_token('ignore', "\t ", states=['metricsconfigurationparams'], include_in_tokens=False)
        self.parser.add_token('newline', func=self.t_newline,  states=['metricsconfigurationparams'], include_in_tokens=False)
        self.parser.add_token('BLOCKCLOSE', func=self.token_metricsconfigurationparams_block_close, states=['metricsconfigurationparams'])
        self.parser.add_token('IDENTIFIER', r'[_a-zA-Z][_a-zA-Z0-9]*', states=['metricsconfigurationparams'])
        self.parser.add_token('SEMICOLON', r'\;', states=['metricsconfigurationparams'])
        self.parser.add_token('EQUAL', func=self.token_metricsconfigurationparams_equal, states=['metricsconfigurationparams'])
        
        # Metrics configuration params value state
        self.parser.add_token('error', func=self.token_error, states=['metricsconfigurationparamsvalue'], include_in_tokens=False)
        self.parser.add_token('ignore', "\t ", states=['metricsconfigurationparamsvalue'], include_in_tokens=False)
        self.parser.add_token('newline', func=self.t_newline,  states=['metricsconfigurationparamsvalue'], include_in_tokens=False)
        self.parser.add_token('PARAMVALUE', func=self.token_metricsconfigurationparamsvalue_paramvalue, states=['metricsconfigurationparamsvalue'])

        # METRICS DATA
        self.parser.add_token('PLUS', r"\+", states=['metricsdata'])
        self.parser.add_token('HASH', r"\#", states=['metricsdata'])
        self.parser.add_token('BLOCKCLOSE', func=self.token_metricsdata_block_close, states=['metricsdata'])

        # Primhead
        self.parser.add_token('SEMICOLON', func=self.token_metricsprimhead_semicolon, states=['metricsprimhead'])
        
        self.parser.add_reserved_word('size', 'SIZE', state='metricsprimhead', case_sensitive=False)
        self.parser.add_reserved_word('exact', 'EXACT', state='metricsprimhead', case_sensitive=False)
        self.parser.add_reserved_word('ratio', 'RATIO', state='metricsprimhead', case_sensitive=False)
        self.parser.add_reserved_word('B', 'B', state='metricsprimhead')
        
        #Primitive
        self.parser.add_token('SQLPARAN', func=self.token_metricsprimitive_sqlparan, states=['metricsprimitive'])
        self.parser.add_token('SEMICOLON', func=self.token_metricsprimitive_semicolon, states=['metricsprimitive'])
        
        self.parser.add_token('error', func=self.token_error, states=['metricsprimitiveargumentvalue'], include_in_tokens=False)
        self.parser.add_token('ignore', "\t ", states=['metricsprimitiveargumentvalue'], include_in_tokens=False)
        self.parser.add_token('newline', func=self.t_newline,  states=['metricsprimitiveargumentvalue'], include_in_tokens=False)
        self.parser.add_token('ARGUMENTVALUE', func=self.token_metricsprimitiveargumentvalue_argumentvalue, states=['metricsprimitiveargumentvalue'])
        
    
        self.parser.add_rule(self.metrics_specification)
        self.parser.add_rule(self.metrics_configurations)
        self.parser.add_rule(self.metrics_configuration)
        self.parser.add_rule(self.metrics_configuration_params)
        self.parser.add_rule(self.metrics_configuration_param)
        self.parser.add_rule(self.metrics_datas)
        self.parser.add_rule(self.metrics_data)
        self.parser.add_rule(self.metrics_data_blocks)
        self.parser.add_rule(self.metrics_data_block)
        self.parser.add_rule(self.metrics_primitives_head)
        self.parser.add_rule(self.metrics_params)
        self.parser.add_rule(self.metrics_param)
        self.parser.add_rule(self.metrics_services_params)
        self.parser.add_rule(self.metrics_primitives)
        self.parser.add_rule(self.metrics_primitive)
        self.parser.add_rule(self.metrics_primitive_arguments)
        self.parser.add_rule(self.metrics_primitive_argument)
        self.parser.add_rule(self.metrics_services_param_size)
        self.parser.add_rule(self.metrics_services_size_type_unit)
        self.parser.add_rule(self.metrics_services_size_type_non_unit)
        self.parser.add_rule(self.metrics_services_size_unit)
    


'''
Created on 01-06-2013

@author: Damian Rusinek <damian.rusinek@gmail.com>
'''

from aqopa.model import MetricsServiceParam
from aqopa.model.parser.lex_yacc import LexYaccParserExtension


class Builder():
    
    def create_metrics_services_param_time(self, token):
        """
        metrics_services_param : SQLPARAN TIME COLON metrics_services_time_type LPARAN metrics_services_time_unit RPARAN SQRPARAN
        """
        return MetricsServiceParam(token[2], token[4], token[6])


class MetricsParserExtension(LexYaccParserExtension):
    """
    Extension for parsing timeanalysis module's metrics
    """
    
    def __init__(self):
        LexYaccParserExtension.__init__(self)
        
        self.builder = Builder()
    

    #######################
    #    Metrics Time
    #######################
    
    def metrics_services_param_time(self, t):
        """
        metrics_services_param : SQLPARAN TIME COLON metrics_services_time_type LPARAN metrics_services_time_unit RPARAN SQRPARAN
        """
        t[0] = self.builder.create_metrics_services_param_time(t)
    
    def metrics_services_time_type(self, t):
        """
        metrics_services_time_type : EXACT
                                | RANGE
        """
        t[0] = t[1].lower()
    
    def metrics_services_time_unit(self, t):
        """
        metrics_services_time_unit : MS
                                | MSPBIT
                                | MSPBYTE
        """
        t[0] = t[1].lower()
    
    def _extend(self):
        """ """
        self.parser.add_reserved_word('time', 'TIME', state='metricsprimhead', case_sensitive=False)
        self.parser.add_reserved_word('exact', 'EXACT', state='metricsprimhead', case_sensitive=False)
        self.parser.add_reserved_word('range', 'RANGE', state='metricsprimhead', case_sensitive=False)
        self.parser.add_reserved_word('ms', 'MS', state='metricsprimhead', case_sensitive=True)
        self.parser.add_reserved_word('mspb', 'MSPBIT', state='metricsprimhead', case_sensitive=True)
        self.parser.add_reserved_word('mspB', 'MSPBYTE', state='metricsprimhead', case_sensitive=True)
        
        self.parser.add_rule(self.metrics_services_param_time)
        self.parser.add_rule(self.metrics_services_time_type)
        self.parser.add_rule(self.metrics_services_time_unit)


'''
Created on 15-05-2013

@author: Damian Rusinek <damian.rusinek@gmail.com>
'''
from aqopa.simulator.state import InstructionsContext,\
    InstructionsList, Process

class Scheduler():
    
    def __init__(self, host):
        raise NotImplementedError()
    
    def finished(self):
        """
        Returns True if host is finished.
        """
        raise NotImplementedError()
    
    def get_current_context(self):
        """
        Returns instructions context executed in current state.
        """
        raise NotImplementedError()
    
    def goto_next_instruction_context(self):
        """
        Move hosts state to next instructions list.
        """
        raise NotImplementedError()

class FifoScheduler(Scheduler):
    
    def __init__(self, host):
        """ """
        self.host = host
        self.context = self._build_context()
    
    def _build_context(self):
        """ """
        context = InstructionsContext()
        context.add_instructions_list(self.host.instructions_list)
        return context
    
    def finished(self):
        """ """
        return self.context.finished()
    
    def get_current_context(self):
        """ """
        return self.context
    
    def goto_next_instruction_context(self):
        """ """
        # Fifo has only one context
        pass
        

class RoundRobinScheduler(Scheduler):
    
    def __init__(self, host):
        """ """
        self.host = host
        self.contexts = []
        self._current_context_index = 0
        self._build_contexts()
    
    def _build_contexts(self):
        """ """
        host_instructions_context = InstructionsContext()
        host_instructions_list = []
        
        host_context_added = False
        for i in self.host.instructions_list:
            
            if isinstance(i, Process):
                process = i
                
                if len(process.instructions_list) > 0:
                    process_context = InstructionsContext()
                    self.contexts.append(process_context)
                    
                    process_context.add_instructions_list(
                                        process.instructions_list, 
                                        process)
            else:
                host_instructions_list.append(i)
                if not host_context_added:
                    self.contexts.append(host_instructions_context)
                    host_context_added = True
                    
        if len(host_instructions_list) > 0:
            host_instructions_context.add_instructions_list(InstructionsList(host_instructions_list))
            
    def finished(self):
        """ """
        for c in self.contexts:
            if not c.finished():
                return False
        return True
    
    def get_current_context(self):
        """ """
        return self.contexts[self._current_context_index]
    
    def goto_next_instruction_context(self):
        """ """
        now_index = self._current_context_index
        
        self._current_context_index = (self._current_context_index + 1) % len(self.contexts)
        
        while now_index != self._current_context_index and self.get_current_context().finished():
            self._current_context_index = (self._current_context_index + 1) % len(self.contexts)
        
        
SCHEDULE_ALGORITHM_ROUND_ROBIN  = 'rr'
SCHEDULE_ALGORITHM_FIFO         = 'fifo'

def create(host, algorithm):
    if algorithm == SCHEDULE_ALGORITHM_FIFO:
        return FifoScheduler(host)
    elif  algorithm == SCHEDULE_ALGORITHM_ROUND_ROBIN:
        return RoundRobinScheduler(host)
    


'''
Created on 22-04-2013

@author: Damian Rusinek <damian.rusinek@gmail.com>
'''

from aqopa.model.parser.lex_yacc import LexYaccParserExtension
from aqopa.model import Function


class Builder():
    """
    Builder for creating function objects
    """
    
    def build_function(self, token):
        """
            FUN IDENTIFIER function_params SEMICOLON 
            FUN IDENTIFIER function_params function_comment SEMICOLON
            FUN IDENTIFIER function_params function_qopml_params SEMICOLON
            FUN IDENTIFIER function_params function_qopml_params function_comment SEMICOLON
        """
        f = Function(token[2], params=token[3])
        
        if len(token) == 6:
            if isinstance(token[4], list):
                f.qop_params = token[4]
            elif isinstance(token[4], str) or isinstance(token[4], unicode):
                f.comment = token[4]
                
        if len(token) == 7:
            f.qop_params = token[4]
            f.comment = token[5]
        
        return f


class ModelParserExtension(LexYaccParserExtension):
    """
    Extension for parsing functions
    """
    
    def __init__(self):
        LexYaccParserExtension.__init__(self)
        
        self.open_blocks_cnt = 0
        self.fun_left_brackets_cnt = 0
        
        self.builder = Builder()

    ##########################################
    #           RESERVED WORDS
    ##########################################
    
    def word_functions_specification(self, t):
        t.lexer.push_state('functions')
        return t
    
    def word_fun(self, t):
        self.fun_left_brackets_cnt = 0
        return t
    
    ##########################################
    #                TOKENS
    ##########################################
    
    def token_block_open(self, t):
        r'{'
        self.open_blocks_cnt += 1
        return t
    
    def token_block_close(self, t):
        r'}'
        self.open_blocks_cnt -= 1
        if self.open_blocks_cnt == 0:
            t.lexer.pop_state()
        return t
    
    def token_lparan(self, t):
        r'\('
        self.fun_left_brackets_cnt += 1
        if self.fun_left_brackets_cnt == 2:
            t.lexer.push_state('functioncomment')
        return t

    # Function Comment State
    def token_funcomment_error(self, t):
        self.parser.syntax_errors.append("Line [%s:%s]: Illegal character '%s' \n" % (t.lexer.lineno, t.lexer.lexpos, t.value[0]))
    
    
    def token_funcomment_rparan(self, t):
        r'\)'
        t.lexer.pop_state()
        return t
    
    def t_newline(self, t):
        r'\n+'
        t.lexer.lineno += t.value.count("\n")
    
    ##########################################
    #                RULES
    ##########################################
    
    def functions_specification(self, t):
        """
        specification : FUNCTIONS_SPECIFICATION BLOCKOPEN functions_list BLOCKCLOSE
        """
        pass
    
    
    def functions_list(self, t):
        """
        functions_list : function 
                    | functions_list function
        """
        pass
    
    def function(self, t):
        """
        function : FUN IDENTIFIER function_params SEMICOLON 
            | FUN IDENTIFIER function_params function_comment SEMICOLON
            | FUN IDENTIFIER function_params function_qopml_params SEMICOLON
            | FUN IDENTIFIER function_params function_qopml_params function_comment SEMICOLON
        """
        self.parser.store.functions.append(self.builder.build_function(t))
    
    def function_comment(self, t):
        """
        function_comment : LPARAN COMMENT RPARAN
        """
        t[0] = t[2]
    
    def function_params(self, t):
        """
        function_params : LPARAN RPARAN
                    | LPARAN identifiers_list RPARAN
        """
        if len(t) > 3:
            t[0] = t[2]
        else:
            t[0] = []

    
    def function_qopml_params(self, t):
        """
        function_qopml_params : SQLPARAN function_qopml_params_list SQRPARAN
        """
        t[0] = t[2]
    
    def function_qopml_params_list(self, t):
        """
        function_qopml_params_list : function_qop_param
                    | function_qopml_params_list function_qop_param
        """
        if len(t) > 2:
            t[0] = t[1] 
            t[0].append(t[2])
        else:
            t[0] = [t[1]]

    def function_qop_param(self, t):
        """
        function_qop_param : IDENTIFIER COLON identifiers_list
        """
        t[0] = (t[1], t[3])
    
    def _extend(self):
        
        self.parser.add_state('functions', 'inclusive')
        self.parser.add_state('functioncomment', 'exclusive')

        self.parser.add_reserved_word('functions', 'FUNCTIONS_SPECIFICATION', func=self.word_functions_specification,)
        self.parser.add_reserved_word('fun', 'FUN', func=self.word_fun, state='functions')

        self.parser.add_token('BLOCKOPEN', func=self.token_block_open, states=['functions'])
        self.parser.add_token('BLOCKCLOSE', func=self.token_block_close, states=['functions'])
        self.parser.add_token('LPARAN', func=self.token_lparan, states=['functions'])
        self.parser.add_token('RPARAN', regex=r'\)', states=['functions'])
        
        # Function comment State
        self.parser.add_token('error', func=self.token_funcomment_error, states=['functioncomment'], include_in_tokens=False)
        self.parser.add_token('ignore', "\t", states=['functioncomment'], include_in_tokens=False)
        self.parser.add_token('newline', func=self.t_newline,  states=['functioncomment'], include_in_tokens=False)
        self.parser.add_token('COMMENT', r'[-_A-Za-z0-9 ]+', states=['functioncomment'])  
        self.parser.add_token('RPARAN', func=self.token_funcomment_rparan, states=['functioncomment'])

        self.parser.add_rule(self.functions_specification)
        self.parser.add_rule(self.functions_list)
        self.parser.add_rule(self.function)
        self.parser.add_rule(self.function_comment)
        self.parser.add_rule(self.function_params)
        self.parser.add_rule(self.function_qopml_params)
        self.parser.add_rule(self.function_qopml_params_list)
        self.parser.add_rule(self.function_qop_param)
        

    


'''
Created on 31-05-2013

@author: Damian Rusinek <damian.rusinek@gmail.com>
'''
import random
from aqopa.simulator.state import Hook, ExecutionResult
from aqopa.model import AssignmentInstruction,\
    CallFunctionInstruction, IfInstruction, WhileInstruction,\
    CommunicationInstruction, CallFunctionExpression, TupleExpression
from aqopa.module.timeanalysis.error import TimeSynchronizationException
from aqopa.simulator.error import RuntimeException

class PreInstructionHook(Hook):
    """
    Execution hook executed before default core execution of each instruction.
    Returns execution result.
    """
    
    def __init__(self, module, simulator):
        """ """
        self.module = module
        self.simulator = simulator
        
        
    def execute(self, context):
        """
        """
        instruction = context.get_current_instruction()
        
        if instruction.__class__ not in [AssignmentInstruction, CallFunctionInstruction, 
                                     IfInstruction, WhileInstruction, CommunicationInstruction]:
            return
        
        if isinstance(instruction, CommunicationInstruction):
            return self._execute_communication_instruction(context)
        else:
            self._update_time(context)
            return ExecutionResult()
            
    def _update_time(self, context):
        """ 
        Update times in context according to current instruction.
        """
        instruction = context.get_current_instruction()
        
        expression = None
        if isinstance(instruction, AssignmentInstruction):
            expression = instruction.expression
        elif isinstance(instruction, CallFunctionInstruction):
            expression = CallFunctionExpression(instruction.function_name, 
                                                instruction.arguments, 
                                                instruction.qop_arguments)
        else:
            expression = instruction.condition
            
        time = self._get_time_for_expression(context, expression)
        
        if time > 0:
            host = context.get_current_host()
            self.module.add_timetrace(self.simulator, host, instruction, 
                                      self.module.get_current_time(self.simulator, host), time)
            self.module.set_current_time(self.simulator, host, self.module.get_current_time(self.simulator, host) + time)
            
    def _get_time_for_expression(self, context, expression):
        """
        Returns calculated time that execution of expression takes.
        """
        if isinstance(expression, TupleExpression):
            return self._calculate_time_for_tuple_expression(context, expression)
        elif isinstance(expression, CallFunctionExpression):
            return self._calculate_time_for_expression(context, expression)
        return 0

    def _calculate_time_for_tuple_expression(self, context, expression):
        """
        Calculate execution time for tuple expression. 
        """
        time = 0
        for i in range(0, len(expression.elements)):
            time += self._get_time_for_expression(context, expression.elements[i])
        return time
    
    def _calculate_time_for_expression(self, context, expression):
        """
        Calculate time for expression.
        """
        time = 0
        metric = context.metrics_manager\
                                .find_primitive(context.get_current_host(), expression)
                                
        if metric:
            block = metric.block
            
            for i in range(0, len(block.service_params)):
                sparam = block.service_params[i]
                
                if sparam.service_name.lower() != "time":
                    continue
                
                metric_type = sparam.param_name.lower()
                metric_unit = sparam.unit
                metric_value = metric.service_arguments[i]
                
                if metric_type == "exact":
                    
                    if metric_unit == "ms":
                        time = float(metric_value)
                        
                    elif metric_unit == "s":
                        time = float(metric_value) * 1000
                        
                    elif metric_unit == "mspb" or metric_unit == "mspB":
                        
                        mparts = metric_value.split(':')
                        if len(mparts) != 2:
                            raise RuntimeException('Metric unit is set as %s, but call parameter to get size of is not set.' 
                                                   % metric_unit)
                        
                        size = 0
                        call_params_indexes = mparts[1].split(',')
                        for index in call_params_indexes:
                            
                            populated_expression = context.expression_populator.populate(
                                                        expression.arguments[index],
                                                        context.get_current_host().get_variables(),
                                                        context.expression_reducer)
                            
                            size += context.metrics_manager.get_expression_size(
                                                                populated_expression,
                                                                context.get_current_host())
                            
                        msperbyte = float(mparts[0])
                        if metric_unit == "mspb":
                            msperbyte = msperbyte / 8.0
                            
                        time = msperbyte * size
                    
                elif metric_type == "range":
                    
                    mvalues = metric_value.split('..')
                    val_from = float(mvalues[0])
                    val_to = float(mvalues[1]) 
                
                    time = val_from + (val_to-val_from)*random.random()
                

        for expr in expression.arguments:
            time += self._get_time_for_expression(context, expr)
            
        return time
                
    def _execute_communication_instruction(self, context):
        """ """
        channel = context.channels_manager.find_channel_for_current_instruction(context)
        if not channel:
            return None
        
        instruction = context.get_current_instruction()
        expressions_cnt = len(instruction.variables_names)
        
        # Check if other hosts should send their message 
        # through this channel before
        current_host_time = self.module.get_current_time(self.simulator, context.get_current_host())
        delay_communication_execution = False
        for h in context.hosts:
            # Omit finished hosts
            if h.finished():
                continue
            # Omit current host
            if h == context.get_current_host():
                continue
            # Omit hosts not using current channel 
            if not channel.connected_with_host(h):
                continue
            
            # Omit hosts that are waiting for the message on current channel
            current_instruction = h.get_current_instructions_context()\
                                    .get_current_instruction()
            if isinstance(current_instruction, CommunicationInstruction):
                
                ch = None
                if instruction.is_out(): # OUT instruction
                    if not current_instruction.is_out(): # IN instruction
                        ch = context.channels_manager.find_channel_for_host_instruction(
                                                        context, h, current_instruction)
                else: # IN instruction
                    if current_instruction.is_out(): # OUT instruction
                        ch = context.channels_manager.find_channel_for_host_instruction(
                                                        context, h, current_instruction)
                if ch == channel:
                    continue
            
            if self.module.get_current_time(self.simulator, h) < current_host_time:
                delay_communication_execution = True
                
        # Delay execution of this instruction 
        # if needed according to previous check      
#        if delay_communication_execution:
#            return ExecutionResult(custom_index_management=True, 
#                                   finish_instruction_execution=True)
                
                
        if instruction.is_out():
            sender = context.get_current_host()
              
            
            receivers_list = channel.get_queue_of_receiving_hosts(expressions_cnt)
            for i in range(0, len(receivers_list)):
                if not receivers_list[i]: # No receiver for message
                    self.module.add_channel_message_trace(self.simulator, channel, 
                                                          self.module.get_channel_next_message_id(self.simulator, channel),
                                                          sender,
                                                          self.module.get_current_time(self.simulator, sender))
                else:
                    if self.module.get_current_time(self.simulator, sender) < self.module.get_current_time(self.simulator, receivers_list[i]):
                        
                        raise TimeSynchronizationException("Time synchronization error. Trying to send message from host '%s' at time %s ms while receiving host '%s' has time %s ms." % 
                                                           (unicode(sender),
                                                            self.module.get_current_time(self.simulator, sender),
                                                            unicode(receivers_list[i]),
                                                            self.module.get_current_time(self.simulator, receivers_list[i])))
                    
                    self.module.add_channel_message_trace(self.simulator, channel,
                                                          self.module.get_channel_next_message_id(self.simulator, channel),
                                                          sender,
                                                          self.module.get_current_time(self.simulator, sender),
                                                          receivers_list[i],
                                                          self.module.get_current_time(self.simulator, receivers_list[i]))
                    self.module.set_current_time(self.simulator, receivers_list[i], self.module.get_current_time(self.simulator, sender))
                
         
        else: # IN communication step
            sender = None
            receiver = context.get_current_host()
            
            sending_hosts = channel.get_queue_of_sending_hosts(expressions_cnt)
            for i in range(0, len(sending_hosts)):
                
                traces = self.module.get_channel_message_traces(self.simulator, channel)
                if len(traces):
                    for j in range(0, len(traces)):
                        if traces[j].sender == sending_hosts[i] and not traces[j].receiver:
                            
                            if self.module.get_current_time(self.simulator, sending_hosts[i]) < self.module.get_current_time(self.simulator, receiver):
                                raise TimeSynchronizationException("Time synchronization error. " + \
                                                           "Trying to send message from host '%s' at time %s ms " + \
                                                           "while receiving host '%s' has time %s ms." % 
                                                           (unicode(sender),
                                                            self.module.get_current_time(self.simulator, sender),
                                                            unicode(receivers_list[i]),
                                                            self.module.get_current_time(self.simulator, receivers_list[i])))
                                
                            traces[j].add_receiver(receiver, self.module.get_current_time(self.simulator, receiver))
                            
            
        


'''
Created on 22-04-2013

@author: Damian Rusinek <damian.rusinek@gmail.com>
'''
from aqopa.model.parser import ParserException, QoPMLModelParser

    
##########################################
#   EXTENDERS 
##########################################

class LexYaccParserExtension():
    """
    Abstract extenstion of parser (tokens, states, rules, etc.)
    """
    
    def __init__(self):
        self.parser = None       # Extended parser
    
    def _save_parser(self, parser):
        if not self.parser:
            self.parser = parser
    
    def _extend(self):
        raise NotImplementedError()
    
    def extend(self, parser):
        self._save_parser(parser)
        self._extend()

##########################################
#   MODEL PARSER 
##########################################

class LexYaccParser(QoPMLModelParser):
    """
    Parser class
    """
    
    def __init__(self):
        self.tokens = ()         # lex tokens
        self.precedence = []     # lex precedence
        self.states = []         # lex states
        self.reserved_words = {} # lex reserved_words
        
        self.start_symbol = None # grammar starting symbol 
        self.lexer = None        # lex object
        self.yaccer = None       # yacc object
        
        self.extensions = []     # objects that extend parser (instances of LexYaccParserExtension)
        
        self.store = None        # store for built objects
        
        self.syntax_errors = []  # list of syntax errors that occured while parsing
        
        
    def get_syntax_errors(self):
        return self.syntax_errors
    
    def add_extension(self, ext):
        self.extensions.append(ext)
        return self
    
    def set_store(self, store):
        self.store = store
        return self
    
    def build(self, **kwargs):
        if not self.lexer:
            
            for e in self.extensions:
                e.extend(self)
            
            # Build the lexer
            from ply import lex
            self.lexer = lex.lex(object=self, **kwargs)
            # Build the yacc
            import ply.yacc as yacc
            self.yaccer = yacc.yacc(module=self, start=self.start_symbol, **kwargs)
        return self
    
    def parse(self, s):
        self.yaccer.parse(input = s, lexer = self.lexer)
        return self.store
        
    # LEX
    
    t_ignore = " \t"
    def t_newline(self, t):
        r'\n+'
        t.lexer.lineno += t.value.count("\n")
        
    def t_error(self, t):
        last_cr = t.lexer.lexdata.rfind('\n',0,t.lexpos)
        if last_cr < 0:
            last_cr = 0
        column = (t.lexpos - last_cr) + 1
        self.syntax_errors.append(("Line [%s:%s, pos:%s]: Illegal character '%s'" % (t.lexer.lineno, column, t.lexer.lexpos, t.value[0])))
        
        next_cr = t.lexer.lexdata.find('\n',t.lexpos)
        if next_cr < 0:
            next_cr = len(t.lexer.lexdata)
        t.lexer.skip(next_cr - t.lexer.lexpos)
        
    def add_state(self, name, state_type):
        if state_type not in ['inclusive', 'exclusive']:
            raise ParserException('Invalid state type State type must be either inclusive or exclusive.')
        self.states.append((name, state_type))
    
    def add_precedence(self, token_name, precedence):
        """
        Adds precedence for token
        """
        self.precedences.append((precedence, token_name))
    
    def add_token(self, name, regex=None, func=None, precedence=None, states=[], include_in_tokens=True):
        """
        Adds new token and saves it in this module. 
        It is needed by 
        """
        
        if include_in_tokens:
            if not name in self.tokens:
                self.tokens += (name,)
        
        field_name = name 
        states_str = '_'.join(states)
        if len(states_str):
            field_name = states_str + '_' + name
        
        if regex and func:
            raise ParserException('Cannot add token with both: regex and function.')
        
        if regex:
            setattr(self, 't_%s' % field_name, regex)
        if func:
            setattr(self, 't_%s' % field_name, func)
            
        if precedence:
            self.add_precedence(name, precedence)
    
    def add_reserved_word(self, word, token, func=None, state='INITIAL', case_sensitive=True):
        """
        Adds token representing reserved word for particular state (INITIAL by default)
        """
        if not token in self.tokens:
            self.tokens += (token,)
        
        if state not in self.reserved_words:
            self.reserved_words[state] = {}
        self.reserved_words[state][word]= (token, func, case_sensitive)
    
    def get_reserved_words(self):
        """
        Returns all reserver words 
        """
        return self.reserved_words
    
    # YACC
    
    def add_rule(self, func):
        setattr(self, 'p_%s' % (func.__name__), func)
    
    def p_error(self, t):
        if not t:
            self.syntax_errors.append("Syntax error 'Unexpected end of file' \n")
        else:
            last_cr = t.lexer.lexdata.rfind('\n',0,t.lexpos)
            if last_cr < 0:
                last_cr = 0
            column = (t.lexpos - last_cr) + 1
            self.syntax_errors.append(("Line [%s:%s, pos: %s]: Syntax error near '%s' \n" % (t.lexer.lineno, column, t.lexer.lexpos, t.value)))
    



'''
Created on 22-04-2013

@author: Damian Rusinek <damian.rusinek@gmail.com>
'''

class QoPMLModelStore():
    
    def __init__(self):
        """ """
        self.functions   = []
        self.channels    = []
        self.equations   = []
        self.versions    = []
        self.hosts       = []
        self.metrics_configurations  = []
        self.metrics_datas           = []
        self.init_version_number     = []
    
    
    def find_host(self, name):
        for h in self.hosts:
            if h.name == name:
                return h
        return None




#!/usr/bin/env python
'''
Created on 05-09-2013

@author: Damian Rusinek <damian.rusinek@gmail.com>
'''
import os
import sys
import time
import threading
import traceback
import wx
import wx.richtext
import wx.lib.newevent
import wx.lib.delayedresult

import aqopa
from aqopa import app
from aqopa.model.parser import MetricsParserException,\
    ConfigurationParserException, ModelParserException
from aqopa.simulator.error import EnvironmentDefinitionException,\
    RuntimeException

ModelParsedEvent, EVT_MODEL_PARSED = wx.lib.newevent.NewEvent()
ModelParseErrorEvent, EVT_MODEL_PARSE_ERROR = wx.lib.newevent.NewEvent()
ModulesChangedEvent, EVT_MODULES_CHANGED = wx.lib.newevent.NewEvent()

# Modules communication events
ModuleSimulationRequestEvent, EVT_MODULE_SIMULATION_REQUEST = wx.lib.newevent.NewEvent() # Parameters: module 
ModuleSimulationAllowedEvent, EVT_MODULE_SIMULATION_ALLOWED = wx.lib.newevent.NewEvent() # Parameters: interpreter
ModuleSimulationFinishedEvent, EVT_MODULE_SIMULATION_FINISHED = wx.lib.newevent.NewEvent()


class ModelPartDataPanel(wx.Panel):
    """ 
    Panel containing text area for one of model parts: 
    model, metrics, configuration.
    """
    
    def __init__(self, *args, **kwargs):
        """ """
        wx.Panel.__init__(self, *args, **kwargs)

        bPanel = wx.Panel(self)
        bSizer = wx.BoxSizer(wx.VERTICAL)
        bPanel.SetSizer(bSizer)

        self.loadButton = wx.Button(bPanel)
        self.saveButton = wx.Button(bPanel)
        
        self.attachButtons(self.loadButton, self.saveButton)
        
        bSizer.Add(self.loadButton, 0, wx.ALL, 5)
        bSizer.Add(self.saveButton, 0, wx.ALL, 5)

        rightPanel = wx.Panel(self)
        rightPanelSizer = wx.BoxSizer(wx.VERTICAL)
        rightPanel.SetSizer(rightPanelSizer)

        self.dataTextArea = wx.richtext.RichTextCtrl(rightPanel, style=wx.TE_MULTILINE | wx.TE_NO_VSCROLL)
        self.dataTextArea.Bind(wx.EVT_KEY_UP, self.printCursorInfo)
        self.dataTextArea.Bind(wx.EVT_LEFT_UP, self.printCursorInfo)
        
        self.cursorInfoLabel = wx.StaticText(rightPanel, label="")
        
        rightPanelSizer.Add(self.dataTextArea, 1, wx.EXPAND)
        rightPanelSizer.Add(self.cursorInfoLabel, 0, wx.ALIGN_RIGHT)

        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(bPanel, 0, wx.ALL, 5)
        sizer.Add(rightPanel, 1, wx.ALL | wx.EXPAND, 5)
        
        self.SetSizer(sizer)
        
        self.Layout()
        
    def attachButtons(self, loadButton, saveButton):
        """ """
        loadButton.Bind(wx.EVT_BUTTON, self.onLoad)
        saveButton.Bind(wx.EVT_BUTTON, self.onSave)
        
    def onLoad(self, event):
        """ Load file to text area """
        ofdlg = wx.FileDialog(self, "Load file", "", "", "QoP-ML Files (*.qopml)|*.qopml", 
                              wx.FD_OPEN | wx.FD_FILE_MUST_EXIST)
        ofdlg.ShowModal()
        if ofdlg.GetPath():
            wildcard, types = wx.richtext.RichTextBuffer.GetExtWildcard(save=False)
            fileType = types[ofdlg.GetFilterIndex()]
            self.dataTextArea.LoadFile(ofdlg.GetPath(), fileType)
        ofdlg.Destroy()
        
    def onSave(self, event):
        """ Save text area value to file """
        ofdlg = wx.FileDialog(self, "Save file", "", "", "QoP-ML Files (*.qopml)|*.qopml", 
                              wx.FD_SAVE)
        ofdlg.ShowModal()
        if ofdlg.GetPath():
            f = open(ofdlg.GetPath(), "w")
            f.write(self.dataTextArea.GetValue())
            f.close()
        ofdlg.Destroy()
        
    def printCursorInfo(self, event):
        """ """
        pos = self.dataTextArea.GetInsertionPoint()
        xy = self.dataTextArea.PositionToXY(pos)
        self.cursorInfoLabel.SetLabel("Line: %d, %d" 
                                      % (xy[1]+1, xy[0]+1))
        self.Layout()
        event.Skip()
        
class ModulesPanel(wx.Panel):
    """ 
    Panel used for selecting modules and configuring them.
    """
    
    def __init__(self, *args, **kwargs):
        self.allModules = kwargs['modules']
        del kwargs['modules']
        
        wx.Panel.__init__(self, *args, **kwargs)
        
        mainSizer = wx.BoxSizer(wx.HORIZONTAL)

        modulesBox = wx.StaticBox(self, label="Modules", size=(100, 100))
        modulesBoxSizer = wx.StaticBoxSizer(modulesBox, wx.VERTICAL)
        
        self.configurationBox = wx.StaticBox(self, label="Configuration", size=(100, 100))
        configurationBoxSizer = wx.StaticBoxSizer(self.configurationBox, wx.VERTICAL)
        
        self.checkBoxesMap = {}
        self.buttonsPanelMap = {}
        self.buttonsModuleGui = {}
        self.modulesPanels = []
        
        emptyPanel = wx.Panel(self, size=(200,20))
        sizer = wx.BoxSizer(wx.VERTICAL)
        text = wx.StaticText(emptyPanel, label="Click 'Configure' button to configure selected module.") 
        sizer.Add(text, 0, wx.ALL | wx.EXPAND, 5)
        emptyPanel.SetSizer(sizer)
        configurationBoxSizer.Add(emptyPanel, 1, wx.ALL | wx.EXPAND, 5)
        self.modulesPanels.append(emptyPanel)

        for m in self.allModules:
            gui = m.get_gui()
            
            modulePanel = wx.Panel(self)
            modulePanelSizer = wx.BoxSizer(wx.HORIZONTAL)
            
            ch = wx.CheckBox(modulePanel, label=gui.get_name())
            ch.Bind(wx.EVT_CHECKBOX, self.OnCheckBoxChange)
            self.checkBoxesMap[m] = ch
            
            btn = wx.Button(modulePanel, label="Configure")
            btn.Bind(wx.EVT_BUTTON, self.OnConfigureButtonClicked)
            
            modulePanelSizer.Add(ch, 0, wx.ALL)
            modulePanelSizer.Add(btn, 0, wx.ALL)
            modulePanel.SetSizer(modulePanelSizer)
            
            modulesBoxSizer.Add(modulePanel, 0, wx.ALL | wx.EXPAND, 5)
            
            moduleConfigurationPanel = gui.get_configuration_panel(self)
            configurationBoxSizer.Add(moduleConfigurationPanel, 1, wx.ALL | wx.EXPAND, 5)
            moduleConfigurationPanel.Hide()

            self.modulesPanels.append(moduleConfigurationPanel)
            self.buttonsPanelMap[btn] = moduleConfigurationPanel
            self.buttonsModuleGui[btn] = gui
            
        mainSizer.Add(modulesBoxSizer, 0, wx.ALL | wx.EXPAND, 5)
        mainSizer.Add(configurationBoxSizer, 1, wx.ALL | wx.EXPAND, 5)
        
        self.SetSizer(mainSizer)
        
    def ShowModuleConfigurationPanel(self, panel):
        """ """
        for p in self.modulesPanels:
            p.Hide()
        panel.Show()
        self.Layout()
        
    def OnCheckBoxChange(self, event):
        """ """
        modules = []
        for m in self.allModules:
            ch = self.checkBoxesMap[m]
            if ch.IsChecked():
                modules.append(m)
        
        wx.PostEvent(self, ModulesChangedEvent(modules = modules))

    def OnConfigureButtonClicked(self, event):
        """ """
        btn = event.EventObject
        moduleGui = self.buttonsModuleGui[btn]
        
        self.configurationBox.SetLabel("%s - Configuration" % moduleGui.get_name())
        self.ShowModuleConfigurationPanel(self.buttonsPanelMap[btn])
        
class RunPanel(wx.Panel):
    """ """
    def __init__(self, *args, **kwargs):
        wx.Panel.__init__(self, *args, **kwargs)
    
        ###############
        # SIMULATION
        ###############    

        self.qopml_model              = ""
        self.qopml_metrics            = ""
        self.qopml_configuration      = ""

        self.selectedModules    = []

        self.interpreter        = None
        self.finishedSimulators = []
        
        self.progressTimer      = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, self.OnProgressTimerTick, self.progressTimer)
        
        ###############
        # LAYOUT
        ###############
        
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        
        leftPanel = wx.Panel(self, style=wx.ALIGN_CENTER)
        self.parseButton = wx.Button(leftPanel, label="Parse")
        self.runButton = wx.Button(leftPanel, label="Run")
        self.runButton.Enable(False)
        panelSizer = wx.BoxSizer(wx.VERTICAL)
        panelSizer.Add(self.parseButton, 0)
        panelSizer.Add(self.runButton, 0)
        leftPanel.SetSizer(panelSizer)
        
        sizer.Add(leftPanel, 0, wx.ALL | wx.EXPAND, 5)
        
        rightPanel = wx.Panel(self)
        self.parsingPanel = self._BuildParsingPanel(rightPanel)
        self.runPanel = self._BuildRunPanel(rightPanel)
        panelSizer = wx.BoxSizer(wx.VERTICAL)
        panelSizer.Add(self.parsingPanel, 1, wx.ALL | wx.EXPAND, 5)
        panelSizer.Add(self.runPanel, 1, wx.ALL | wx.EXPAND, 5)
        self.runPanel.Hide()
        rightPanel.SetSizer(panelSizer)

        sizer.Add(rightPanel, 1, wx.ALL | wx.EXPAND, 5)
        
        self.SetSizer(sizer)
        
        ###############
        # EVENTS
        ###############
        
        self.parseButton.Bind(wx.EVT_BUTTON, self.OnParseClicked)
        self.runButton.Bind(wx.EVT_BUTTON, self.OnRunClicked)
        
        self.Bind(EVT_MODEL_PARSED, self.OnModelParsed)
        
    def _BuildParsingPanel(self, parent):
        """ """
        panel = wx.Panel(parent, style=wx.ALIGN_CENTER)
        sizer = wx.BoxSizer(wx.VERTICAL)
        
        text = wx.StaticText(panel, label="Parsing Info")
        self.parseResult = wx.TextCtrl(panel, 
                                     style=wx.TE_MULTILINE | wx.TE_READONLY)
        
        sizer.Add(text, 0, wx.ALL|wx.EXPAND, 5)
        sizer.Add(self.parseResult, 1, wx.ALL|wx.EXPAND, 5)
    
        panel.SetSizer(sizer)
        return panel
    
    def _BuildRunPanel(self, parent):
        """ """
        
        panel = wx.Panel(parent)
        sizer = wx.BoxSizer(wx.VERTICAL)
        
        statusStaticBox = wx.StaticBox(panel, label="Status")
        statusStaticBoxSizer = wx.StaticBoxSizer(statusStaticBox, wx.VERTICAL)

        self.statusLabel = wx.StaticText(panel, label="Running")
        self.percentLabel = wx.StaticText(panel, label="0%")
        self.dotsLabel = wx.StaticText(panel, label=".")
        
        statusStaticBoxSizer.Add(self.statusLabel, 0, wx.ALL|wx.ALIGN_CENTER, 5)
        statusStaticBoxSizer.Add(self.percentLabel, 0, wx.ALL|wx.ALIGN_CENTER, 5)
        statusStaticBoxSizer.Add(self.dotsLabel, 0, wx.ALL|wx.ALIGN_CENTER, 5)
        
        
        timeStaticBox = wx.StaticBox(panel, label="Analysis Time")
        timeStaticBoxSizer = wx.StaticBoxSizer(timeStaticBox, wx.VERTICAL)
        
        self.analysisTime = wx.StaticText(panel, label='---')
        
        timeStaticBoxSizer.Add(self.analysisTime, 0, wx.ALL | wx.ALIGN_CENTER, 5)
        
        text = wx.StaticText(panel, label="Run Info")
        self.runResult = wx.TextCtrl(panel, 
                                     style=wx.TE_MULTILINE | wx.TE_READONLY)
    
        sizer.Add(statusStaticBoxSizer, 0, wx.ALL | wx.EXPAND, 5)
        sizer.Add(timeStaticBoxSizer, 0, wx.ALL | wx.EXPAND, 5)    
        sizer.Add(text, 0, wx.ALL|wx.EXPAND, 5)
        sizer.Add(self.runResult, 1, wx.ALL|wx.EXPAND, 5)
    
    
        panel.SetSizer(sizer)
        return panel
    
    def ShowPanel(self, panel):
        self.parsingPanel.Hide()
        self.runPanel.Hide()
        panel.Show()
        self.Layout()
        
    def SetModel(self, model, metrics, configuration):
        """ """
        self.qopml_model              = model
        self.qopml_metrics            = metrics
        self.qopml_configuration      = configuration
        
    def SetSelectedModules(self, modules):
        """ """
        self.selectedModules = modules
        
    def OnParseClicked(self, event):
        """ """
        self.interpreter = app.GuiInterpreter(
                                     model_as_text=self.qopml_model, 
                                     metrics_as_text=self.qopml_metrics,
                                     config_as_text=self.qopml_configuration)
        
        for m in self.selectedModules:
            self.interpreter.register_qopml_module(m)
        
        try:
            resultMessage = ""
            error = False
            self.interpreter.parse()
            resultMessage = "SUCCESFULLY PARSED\n\n Now you can run simulation."
            wx.PostEvent(self, ModelParsedEvent())
        except EnvironmentDefinitionException, e:
            error = True
            resultMessage = "ENVIRONMENT ERROR\n"
            resultMessage += "%s\n" % unicode(e)
        except ModelParserException, e:
            error = True
            resultMessage = "MODEL SYNTAX ERROR\n"
            if len(e.syntax_errors):
                resultMessage += "\n".join(e.syntax_errors)
        except MetricsParserException, e:
            error = True
            resultMessage = "METRICS SYNTAX ERROR\n"
            if len(e.syntax_errors):
                resultMessage += "\n".join(e.syntax_errors)
        except ConfigurationParserException, e:
            error = True
            resultMessage = "VERSIONS SYNTAX ERROR\n"
            if len(e.syntax_errors):
                resultMessage += "\n".join(e.syntax_errors)
        
        if error:
            resultMessage += "\nModel may include syntax parsed by modules (eg. metrics, configuration). "+\
                             "Have you selected modules?"
            wx.PostEvent(self, ModelParseErrorEvent(error=resultMessage))
        if resultMessage != "":
            self.parseResult.SetValue(resultMessage)
            
        self.ShowPanel(self.parsingPanel)
            
    def OnModelParsed(self, event):
        """ """
        self.runButton.Enable(True)
            
    def OnRunClicked(self, event):
        """ """
        try:
            self.statusLabel.SetLabel("Running")
            self.analysisTime.SetLabel("---")
            self.percentLabel.SetLabel("0%")
            self.runResult.SetValue("")
            
            self.runButton.Enable(False)
            self.ShowPanel(self.runPanel)
            
            self.finishedSimulators = []
            self.simulatorIndex = 0
            
            self.startAnalysisTime = time.time()
            
            self.interpreter.prepare()
            self.progressTimer.Start(1000)

            simulator = self.interpreter.simulators[self.simulatorIndex]
            wx.lib.delayedresult.startWorker(self.OnSimulationFinished, 
                                             self.interpreter.run_simulation, 
                                             wargs=(simulator,),
                                             jobID = self.simulatorIndex)
                
            
        except EnvironmentDefinitionException, e:
            self.statusLabel.SetLabel("Error")
            self.runButton.Enable(True)
            errorMessage = "Error on creating environment: %s\n" % e
            if len(e.errors) > 0:
                errorMessage += "Errors:\n"
                errorMessage += "\n".join(e.errors)
            self.runResult.SetValue(errorMessage)
            self.progressTimer.Stop()
            
        except Exception, e:
            self.statusLabel.SetLabel("Error")
            self.runButton.Enable(True)
            sys.stderr.write(traceback.format_exc())
            errorMessage = "Unknown error\n"
            self.runResult.SetValue(errorMessage)
            self.progressTimer.Stop()
        
        self.ShowPanel(self.runPanel)
        
    def OnSimulationFinished(self, result):
        """ """
        simulator = self.interpreter.simulators[self.simulatorIndex]
        self.simulatorIndex += 1
        self.finishedSimulators.append(simulator)
        
        resultMessage = None
        try :
            simulator = result.get()
    
            self.PrintProgressbar(self.GetProgress())
            
            for m in self.selectedModules:
                gui = m.get_gui()
                gui.on_finished_simulation(simulator)
                
            runResultValue = self.runResult.GetValue()
            resultMessage = runResultValue + \
                "Version %s finished successfully.\n" \
                % simulator.context.version.name
                    
        except RuntimeException, e:
            runResultValue = self.runResult.GetValue()
            resultMessage = runResultValue + \
                "Version %s finished with error: %s.\n" \
                % (simulator.context.version.name, e.args[0])
        except Exception, e:
            sys.stderr.write(traceback.format_exc())
            runResultValue = self.runResult.GetValue()
            resultMessage = runResultValue + \
                "Version %s finished with unknown error.\n" \
                % simulator.context.version.name
        
        if resultMessage:    
            self.runResult.SetValue(resultMessage)
        
        if len(self.finishedSimulators) == len(self.interpreter.simulators):
            self.OnAllSimulationsFinished()
        else:
            simulator = self.interpreter.simulators[self.simulatorIndex]
            wx.lib.delayedresult.startWorker(self.OnSimulationFinished, 
                                             self.interpreter.run_simulation, 
                                             wargs=(simulator,),
                                             jobID = self.simulatorIndex)
                
    def OnAllSimulationsFinished(self):
        """ """
        self.progressTimer.Stop()
        self.PrintProgressbar(1)
        
        self.endAnalysisTime = time.time()
        timeDelta = self.endAnalysisTime - self.startAnalysisTime
        analysisTimeLabel = "%.4f s" % timeDelta 
        
        self.analysisTime.SetLabel(analysisTimeLabel)
        self.Layout()
        
        for m in self.selectedModules:
            gui = m.get_gui()
            gui.on_finished_all_simulations(self.interpreter.simulators)
            
    ################
    # PROGRESS BAR 
    ################
        
    def OnProgressTimerTick(self, event):
        """ """
        progress = self.GetProgress()
        if progress == 1:
            self.progressTimer.Stop()
        self.PrintProgressbar(progress)
        
    def GetProgress(self):
        all = 0.0
        sum = 0.0
        for simulator in self.interpreter.simulators:
            all += 1
            sum += simulator.context.get_progress()
        progress = 0
        if all > 0:
            progress = sum / all
        return progress
        
    def PrintProgressbar(self, progress):
        """
        Prints the formatted progressbar showing the progress of simulation. 
        """
        percentage = str(int(round(progress*100))) + '%'
        self.percentLabel.SetLabel(percentage)
        self.runPanel.Layout()
        
        if progress == 1:
            self.statusLabel.SetLabel('Finished')
            self.runPanel.Layout()
            self.dotsLabel.SetLabel('')
            self.runPanel.Layout()
        else:
            dots = self.dotsLabel.GetLabel()
            if len(dots) > 10:
                dots = "."
            else:
                dots += ".."
            self.dotsLabel.SetLabel(dots)
            self.runPanel.Layout()

        
class ResultsPanel(wx.Panel):
    """ """
    def __init__(self, *args, **kwargs):
        wx.Panel.__init__(self, *args, **kwargs)
        
        self.selectedModules = []
        self.moduleResultPanel = {}
        self.buttonsModule = {}
        
        self._BuildMainLayout()
        
    def _BuildMainLayout(self):
        
        mainSizer = wx.BoxSizer(wx.HORIZONTAL)

        self.modulesBox = wx.StaticBox(self, label="Modules", size=(100, 100))
        self.modulesBox.Hide()
        self.modulesBoxSizer = wx.StaticBoxSizer(self.modulesBox, wx.VERTICAL)
        
        self.resultsBox = wx.StaticBox(self, label="Results", size=(100, 100))
        self.resultsBox.Hide()
        self.resultsBoxSizer = wx.StaticBoxSizer(self.resultsBox, wx.VERTICAL)
        
        mainSizer.Add(self.modulesBoxSizer, 0, wx.ALL | wx.EXPAND, 5)
        mainSizer.Add(self.resultsBoxSizer, 1, wx.ALL | wx.EXPAND, 5)
        
        self.SetSizer(mainSizer)
        self.Layout()
        
    def _BuildModulesLayout(self):
        """ """
        
        for m in self.selectedModules:
            if m in self.moduleResultPanel:
                continue
            
            gui = m.get_gui()
            
            btn = wx.Button(self, label=gui.get_name())
            btn.Bind(wx.EVT_BUTTON, self.OnModuleButtonClicked)
            self.modulesBoxSizer.Add(btn, 0, wx.ALL | wx.EXPAND)
            self.buttonsModule[btn] = m
            
            resultPanel = gui.get_results_panel(self)
            self.resultsBoxSizer.Add(resultPanel, 1, wx.ALL | wx.EXPAND)
            self.moduleResultPanel[m] = resultPanel
            
            self.Layout()
#            resultPanel.Hide()
            
        uncheckedModules = []
        for m in self.moduleResultPanel:
            if m not in self.selectedModules:
                uncheckedModules.append(m)

        buttonsToRemove = []
        for m in uncheckedModules:
            self.moduleResultPanel[m].Destroy()
            del self.moduleResultPanel[m]
            
            for btn in self.buttonsModule:
                if self.buttonsModule[btn] == m:
                    buttonsToRemove.append(btn)
                    
        for btn in buttonsToRemove:
            btn.Destroy()
            del self.buttonsModule[btn]
            
        self.Layout()
        
    def SetSelectedModules(self, modules):
        """ """
        self.selectedModules = modules
        
        if len(self.selectedModules) > 0:
            self.modulesBox.Show()
            self.resultsBox.Show()
        else:
            self.modulesBox.Hide()
            self.resultsBox.Hide()
        
        self._BuildModulesLayout()
        
    def ClearResults(self):
        """ """
        for m in self.selectedModules:
            gui = m.get_gui()
            gui.on_parsed_model()
                    
    def OnModuleButtonClicked(self, event):
        """ """
        btn = event.EventObject
        
class MainNotebook(wx.Notebook):
    """ """
    
    def __init__(self, parent):
        wx.Notebook.__init__(self, parent)
        
        ###########
        # MODULES 
        ###########
        
        self.availableModules = []
        
        from aqopa.module import timeanalysis
        m = timeanalysis.Module()
        m.get_gui().Bind(EVT_MODULE_SIMULATION_REQUEST, self.OnModuleSimulationRequest)
        m.get_gui().Bind(EVT_MODULE_SIMULATION_FINISHED, self.OnModuleSimulationFinished)
        self.availableModules.append(m)

        ###########
        # TABS
        ###########

        self.modelTab = ModelPartDataPanel(self)
        self.modelTab.loadButton.SetLabel("Load Model")
        self.modelTab.saveButton.SetLabel("Save Model")
        self.modelTab.Layout()
        self.Bind(wx.EVT_TEXT, self.OnModelTextChange, self.modelTab.dataTextArea)
        self.AddPage(self.modelTab, "Model")
        
        self.metricsTab = ModelPartDataPanel(self)
        self.metricsTab.loadButton.SetLabel("Load Metrics")
        self.metricsTab.saveButton.SetLabel("Save Metrics")
        self.metricsTab.Layout()
        self.Bind(wx.EVT_TEXT, self.OnModelTextChange, self.metricsTab.dataTextArea)
        self.AddPage(self.metricsTab, "Metrics")
        
        self.configurationTab = ModelPartDataPanel(self)
        self.configurationTab.loadButton.SetLabel("Load Versions")
        self.configurationTab.saveButton.SetLabel("Save Versions")
        self.configurationTab.Layout()
        self.Bind(wx.EVT_TEXT, self.OnModelTextChange, self.configurationTab.dataTextArea)
        self.configurationTab.Layout()
        self.AddPage(self.configurationTab, "Versions")
        
        self.modulesTab = ModulesPanel(self, modules=self.availableModules)
        self.modulesTab.Bind(EVT_MODULES_CHANGED, self.OnModulesChange)
        self.modulesTab.Layout()
        self.AddPage(self.modulesTab, "Modules")
        
        self.runTab = RunPanel(self)
        self.runTab.Layout()
        self.runTab.Bind(EVT_MODEL_PARSED, self.OnModelParsed)
        self.AddPage(self.runTab, "Run")
        
        self.resultsTab = ResultsPanel(self)
        self.resultsTab.Layout()
        self.AddPage(self.resultsTab, "Results")
        
        
    def LoadModelFile(self, filePath):
        self.modelTab.dataTextArea.LoadFile(filePath)
        
    def LoadMetricsFile(self, filePath):
        self.metricsTab.dataTextArea.LoadFile(filePath)
        
    def LoadVersionsFile(self, filePath):
        self.configurationTab.dataTextArea.LoadFile(filePath)
        
    def SetModelData(self, data):
        self.modelTab.dataTextArea.SetValue(data)
        
    def SetMetricsData(self, data):
        self.metricsTab.dataTextArea.SetValue(data)
        
    def SetVersionsData(self, data):
        self.configurationTab.dataTextArea.SetValue(data)
        
    def GetModelData(self):
        return self.modelTab.dataTextArea.GetValue().strip()
        
    def GetMetricsData(self):
        return self.metricsTab.dataTextArea.GetValue().strip()
        
    def GetVersionsData(self):
        return self.configurationTab.dataTextArea.GetValue().strip()
        
    def OnModelTextChange(self, event):
        self.runTab.SetModel(self.GetModelData(), 
                             self.GetMetricsData(),
                             self.GetVersionsData())
        event.Skip()
        
    def OnModulesChange(self, event):
        self.runTab.SetSelectedModules(event.modules)
        self.resultsTab.SetSelectedModules(event.modules)
        
    def OnModelParsed(self, event):
        self.resultsTab.ClearResults()
        event.Skip()
        
    def OnModuleSimulationRequest(self, event):
        """ """
        gui = event.module.get_gui()
        self.runTab.runButton.Enable(False)
        self.runTab.parseButton.Enable(False)
        
        wx.PostEvent(gui, ModuleSimulationAllowedEvent(interpreter=self.runTab.interpreter))

    def OnModuleSimulationFinished(self, event):
        """ """
        self.runTab.parseButton.Enable(True)

######################################
#            LIBRARY
######################################

ModelSelectedEvent, EVT_MODEL_SELECTED = wx.lib.newevent.NewEvent()
        
class LibraryTree(wx.TreeCtrl):
    """ """
    def __init__(self, *args, **kwargs):
        wx.TreeCtrl.__init__(self, *args, **kwargs)

        models_item = self.AddRoot(text="Models")
        models_dir = os.path.join(os.path.dirname(__file__),
                                  os.pardir,
                                  'library', 
                                  'models')
        
        import xml.etree.ElementTree as ET
        
        for dir_root, dirs, files in os.walk(models_dir):
            if 'meta.xml' in files:
                item_key = os.path.basename(dir_root)
                
                tree = ET.parse(os.path.join(dir_root, 'meta.xml'))
                root = tree.getroot()
                
                name_child = root.find('name')
                if name_child is not None: 
                    item = self.AppendItem(models_item, text=name_child.text)
                    
                    author = root.find('author').text if root.find('author') is not None else ''
                    author_email = root.find('author_email').text if root.find('author_email') is not None else ''
                    description = root.find('description').text if root.find('description') is not None else ''
                    
                    model_file = ''
                    metrics_file = ''
                    versions_file = ''
                    files = root.find('files')
                    if files is not None:
                        model_file = files.find('model').text if files.find('model') is not None else ''
                        metrics_file = files.find('metrics').text if files.find('metrics') is not None else ''
                        versions_file = files.find('versions').text if files.find('versions') is not None else '' 
                    
                    model_data = {
                        'root':         dir_root,
                        'name':         name_child.text,
                        'author':       author,
                        'author_email': author_email,
                        'description':  description,
                        
                        'files': {
                            'model':    model_file,
                            'metrics':  metrics_file,
                            'versions': versions_file,
                        }
                    }
                    self.SetPyData(item, model_data)
        self.ExpandAll()
        
    def GetModelData(self, item):
        """ """
        if item in self.items_data:
            return self.items_data[item]
        return None
                    
class ModelDescriptionPanel(wx.Panel):
    """ """
    def __init__(self, *args, **kwargs):
        wx.Panel.__init__(self, *args, **kwargs)
        
        self.model_data = None
        
        self.modelDescriptionBox = wx.StaticBox(self, label="Model Description")
        modelDescriptionBoxSizer = wx.StaticBoxSizer(self.modelDescriptionBox, wx.VERTICAL)
        
        self.nameText = wx.StaticText(self, label="Module name")
        nameFont = wx.Font(14, wx.MODERN, wx.NORMAL, wx.NORMAL)
        self.nameText.SetFont(nameFont)
        modelDescriptionBoxSizer.Add(self.nameText, 0, wx.ALIGN_CENTER)
        
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        authorLabelText = wx.StaticText(self, label="Author:")
        sizer.Add(authorLabelText, 0, wx.ALL, 5)
        self.authorText = wx.StaticText(self)
        sizer.Add(self.authorText, 0, wx.ALL, 5)
        modelDescriptionBoxSizer.Add(sizer, 0, wx.ALIGN_CENTER)
        
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        authorEmailLabelText = wx.StaticText(self, label="Author e-mail:")
        sizer.Add(authorEmailLabelText, 0, wx.ALL, 5)
        self.authorEmailText = wx.StaticText(self)
        sizer.Add(self.authorEmailText, 0, wx.ALL, 5)
        modelDescriptionBoxSizer.Add(sizer, 0, wx.ALIGN_CENTER)
        
        self.descriptionText = wx.TextCtrl(self, style = wx.TE_MULTILINE|wx.TE_READONLY|wx.TE_AUTO_URL)
        self.descriptionText.Hide()
        modelDescriptionBoxSizer.Add(self.descriptionText, 1, wx.EXPAND)
        
        self.loadModelBtn = wx.Button(self, label="Load model")
        self.loadModelBtn.Bind(wx.EVT_LEFT_UP, self.OnLoadModelClicked)
        self.loadModelBtn.Hide()
        modelDescriptionBoxSizer.Add(self.loadModelBtn, 0, wx.EXPAND)
        
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(modelDescriptionBoxSizer, 1, wx.EXPAND)
        self.SetSizer(sizer)
        self.Layout()
        
    def ShowModel(self, model_data):
        """ """
        self.model_data = model_data
        self.nameText.SetLabel(model_data['name'])
        self.authorText.SetLabel(model_data['author'])
        self.authorEmailText.SetLabel(model_data['author_email'])
        self.descriptionText.SetValue(model_data['description'])
        self.descriptionText.Show()
        self.loadModelBtn.Show()
        self.Layout()
        
    def OnLoadModelClicked(self, event=None):
        """ """
        f = open(os.path.join(self.model_data['root'], self.model_data['files']['model']))
        model_data = f.read()
        f.close()
        
        f = open(os.path.join(self.model_data['root'], self.model_data['files']['metrics']))
        metrics_data = f.read()
        f.close()
        
        f = open(os.path.join(self.model_data['root'], self.model_data['files']['versions']))
        versions_data = f.read()
        f.close() 
        
        evt = ModelSelectedEvent(model_data=model_data,
                                 metrics_data=metrics_data,
                                 versions_data=versions_data)
        wx.PostEvent(self, evt)
        
class LibraryFrame(wx.Frame):
    """ """
    def __init__(self, *args, **kwargs):
        wx.Frame.__init__(self, *args, **kwargs)
        
        ###################
        # SIZERS & EVENTS
        ###################
        
        self.modelsTree = LibraryTree(self)
        self.modelDescriptionPanel = ModelDescriptionPanel(self)
        self.modelDescriptionPanel.Bind(EVT_MODEL_SELECTED, self.OnLoadModelSelected)
        
        self.modelsTree.Bind(wx.EVT_TREE_SEL_CHANGED, self.OnModelSelected)
        self.modelsTree.Bind(wx.EVT_TREE_ITEM_ACTIVATED, self.OnModelDoubleClicked)
        
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(self.modelsTree, 1, wx.EXPAND)
        sizer.Add(self.modelDescriptionPanel, 4, wx.EXPAND)
        self.SetSizer(sizer)
        
        self.Layout()
        
    def OnModelSelected(self, event=None):
        """ """
        itemID = event.GetItem()
        if not itemID.IsOk():
            itemID = self.tree.GetSelection()
        model_data = self.modelsTree.GetPyData(itemID)
        if model_data:
            self.modelDescriptionPanel.ShowModel(model_data)

    def OnModelDoubleClicked(self, event=None):
        """ """
        itemID = event.GetItem()
        if not itemID.IsOk():
            itemID = self.tree.GetSelection()
        model_info = self.modelsTree.GetPyData(itemID)
        if model_info:
            f = open(os.path.join(model_info['root'], model_info['files']['model']))
            model_data = f.read()
            f.close()

            f = open(os.path.join(model_info['root'], model_info['files']['metrics']))
            metrics_data = f.read()
            f.close()

            f = open(os.path.join(model_info['root'], model_info['files']['versions']))
            versions_data = f.read()
            f.close()

            evt = ModelSelectedEvent(model_data=model_data,
                                     metrics_data=metrics_data,
                                     versions_data=versions_data)
            wx.PostEvent(self, evt)
            self.Close()

    def OnLoadModelSelected(self, event=None):
        """ """
        wx.PostEvent(self, event)
        self.Close()
        
class MainFrame(wx.Frame):
    """ """
    def __init__(self, *args, **kwargs):
        wx.Frame.__init__(self, *args, **kwargs)

        ###########
        # MENUBAR 
        ###########        
        
        menuBar = wx.MenuBar()

        fileMenu = wx.Menu()
        item = fileMenu.Append(wx.ID_ABOUT, text="About AQoPA")
        self.Bind(wx.EVT_MENU, self.OnAbout, item)
        fileMenu.AppendSeparator()
        item = fileMenu.Append(wx.ID_EXIT, text="&Quit")
        self.Bind(wx.EVT_MENU, self.OnQuit, item)
        menuBar.Append(fileMenu, "&File")
        
        libraryMenu = wx.Menu()
        item = libraryMenu.Append(-1, text="Browse models")
        self.Bind(wx.EVT_MENU, self.OnBrowseModels, item)
        menuBar.Append(libraryMenu, "&Library")
        
        self.SetMenuBar(menuBar)
        
        ###################
        # SIZERS & EVENTS
        ###################
        
        self.mainNotebook = MainNotebook(self)
        
        logo_filepath = os.path.join(os.path.dirname(__file__), 
                                     'assets', 
                                     'logo.png')
        logoPanel = wx.Panel(self)
        pic = wx.StaticBitmap(logoPanel)
        pic.SetBitmap(wx.Bitmap(logo_filepath))
        
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(logoPanel, 0, wx.CENTER, 5)
        sizer.Add(self.mainNotebook, 1, wx.ALL|wx.EXPAND, 5)
        self.SetSizer(sizer)
        
        self.Layout()
        
    def OnQuit(self, event=None):
        """ Close app """
        self.Close()
        
    def OnBrowseModels(self, event=None):
        """ Show frame with library """
        libraryFrame = LibraryFrame(self, title="Models Library")
        libraryFrame.Show(True)
        libraryFrame.Maximize(True)
        
        libraryFrame.Bind(EVT_MODEL_SELECTED, self.OnLibraryModelSelected)
        
    def OnLibraryModelSelected(self, event):
        """ """
        self.mainNotebook.SetModelData(event.model_data)
        self.mainNotebook.SetMetricsData(event.metrics_data)
        self.mainNotebook.SetVersionsData(event.versions_data) 
        
    def OnAbout(self, event=None):
        """ Show about info """
        
        description = """AQoPA stands for Automated Quality of Protection Analysis Tool 
        for QoPML models."""

        licence = """AQoPA is free software; you can redistribute 
it and/or modify it under the terms of the GNU General Public License as 
published by the Free Software Foundation; either version 2 of the License, 
or (at your option) any later version.

AQoPA is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of 
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE."""
        
        logo_filepath = os.path.join(os.path.dirname(__file__), 
                                     'assets', 
                                     'logo.png')
        
        info = wx.AboutDialogInfo()
        info.SetIcon(wx.Icon(logo_filepath, wx.BITMAP_TYPE_PNG))
        info.SetName('AQoPA')
        info.SetVersion(aqopa.VERSION)
        info.SetDescription(description)
        info.SetCopyright('(C) 2013 QoPML Project')
        info.SetWebSite('http://www.qopml.org')
        info.SetLicence(licence)
        info.AddDeveloper('Damian Rusinek')
        info.AddDocWriter('Damian Rusinek')
        info.AddArtist('QoPML Project')
        info.AddTranslator('Damian Rusinek')
        
        wx.AboutBox(info)
        
        
class AqopaApp(wx.App):
    
    def OnInit(self):
        self.mainFrame = MainFrame(None, 
                                   title="Automated Quality of Protection Analysis Tool")
        self.mainFrame.Show(True)
        self.mainFrame.Maximize(True)
        self.SetTopWindow(self.mainFrame)
        
        return True
    



'''
Created on 07-05-2013

@author: Damian Rusinek <damian.rusinek@gmail.com>
'''

import wx
import wx.lib.newevent
import threading
from aqopa.model.parser import ParserException, ModelParserException,\
    MetricsParserException, ConfigurationParserException
from aqopa.model.store import QoPMLModelStore
from aqopa.model import HostProcess, name_indexes, original_name,\
    HostSubprocess, WhileInstruction, IfInstruction
from aqopa.simulator import Simulator,\
    expression, state, equation, metrics, communication, scheduler
    
from aqopa.simulator.state import Executor,\
    AssignmentInstructionExecutor, IfInstructionExecutor,\
    ProcessInstructionExecutor, SubprocessInstructionExecutor,\
    FinishInstructionExecutor, CommunicationInstructionExecutor,\
    ContinueInstructionExecutor, WhileInstructionExecutor, Host, Process,\
    CallFunctionInstructionExecutor, PrintExecutor
    
from aqopa.simulator.error import EnvironmentDefinitionException

class Builder():
    """
    Builder that builds environment elements
    """
    
    def build_store(self):
        """
        Builds store - keeps all parsed elements from qopml model
        """
        return QoPMLModelStore()
    
    def _build_functions(self, store):
        """
        Validate and build functions
        """
        functions = []
        errors = []
        
        for f in store.functions:
            err = False
            for ef in functions:
                if ef.name == f.name:
                    errors.append("Function %s redeclared" % f.name)
                    err = True
                    break
            if not err:
                functions.append(f)
                
        if len(errors) > 0:
            raise EnvironmentDefinitionException('Functions redeclaration.', errors=errors)
        
        return functions
    
    def _build_equations(self, store, functions):
        """
        Validate, build and return simulation equations build from parsed equations.
        """
        validator = equation.Validator()
        validator.validate(store.equations, functions)
        
        equations = []
        for parsed_equation in store.equations:
            equations.append(equation.Equation(parsed_equation.composite, parsed_equation.simple))
        return equations
    
    def _build_hosts(self, store, version, functions, populator, reducer):
        """
        Rebuild hosts - create new instances with updated instructions lists according to version.
        """
        
        def remove_unused_subprocesses(instructions_list, run_process): 
            """
            Create instruction lists according to "run process"
            Removes subprocesses that are not selected to run. 
            """
            if run_process.all_subprocesses_active:
                return instructions_list
            
            new_instructions_list = []
            
            for instruction in instructions_list:
                if isinstance(instruction, HostSubprocess):
                    subprocess = instruction
                    if subprocess.name in run_process.active_subprocesses:
                        new_instructions_list.append(subprocess)
                        subprocess.instructions_list = \
                            remove_unused_subprocesses(subprocess.instructions_list, 
                                                       run_process)
                else:
                    new_instructions_list.append(instruction)
                    
                    if isinstance(instruction, WhileInstruction):
                        instruction.instructions = \
                            remove_unused_subprocesses(instruction.instructions, 
                                                       run_process)
                    
                    if isinstance(instruction, IfInstruction):
                        instruction.true_instructions = \
                            remove_unused_subprocesses(instruction.true_instructions, 
                                                       run_process)
                        instruction.false_instructions = \
                            remove_unused_subprocesses(instruction.false_instructions, 
                                                       run_process)
            
            return new_instructions_list
            
        def build_host_instructions_list(parsed_host, run_host, repetition_number):
            """
            Create host's  and its processes' instruction lists 
            according to "run host" (repetitions) 
            """
            
            def find_process(instructions_list, process_name):
                for instr in instructions_list:
                    if not isinstance(instr, HostProcess):
                        continue
                    if instr.name == process_name:
                        return instr
                return None
            
            
            processes_numbers = {}
            host_instructions_list = []
            
            for run_process in run_host.run_processes:
                
                # Find parsed process
                parsed_process = find_process(parsed_host.instructions_list, run_process.process_name)
                
                if parsed_process is None:
                    raise EnvironmentDefinitionException("Process '%s' does not exist in host '%s'." % 
                                                         (run_process.process_name, parsed_host.name))
                
                # Clone the parsed process, because builder 
                # may change its the instructions list
                # and remove some subprocesses that would be 
                # needed by other version 
                parsed_process = parsed_process.clone()
                
                # Define initial process number (if needed)
                if run_process.process_name not in processes_numbers:
                    processes_numbers[run_process.process_name] = 0
                
                # If process has follower
                if run_process.follower:
                    # Define initial process number for following process (if needed)
                    if run_process.follower.process_name not in processes_numbers:
                        processes_numbers[run_process.follower.process_name] = 0
                    
                for i in range(0, run_process.repetitions):
                    
                    # Create instructions list
                    instructions_list = remove_unused_subprocesses(parsed_process.instructions_list, run_process)
                    
                    # Create new simulation process
                    simulated_process = Process(parsed_process.name, instructions_list)
                    
                    # Update process index
                    process_number = processes_numbers[run_process.process_name] + i
                    simulated_process.add_name_index(process_number)
                    
                    # Do the same fo the follower 
                    if run_process.follower:
                        
                        # Find process of follower
                        follower_parsed_process = find_process(parsed_host.instructions_list, run_process.follower.process_name)
                        
                        # Build instructions list for follower
                        follower_instructions_list = remove_unused_subprocesses(follower_parsed_process.instructions_list, run_process.follower)

                        # Create simulated follower
                        simulated_follower = Process(follower_parsed_process.name, follower_instructions_list)
                            
                        # Update follower index
                        follower_number = processes_numbers[run_process.follower.process_name] + i
                        simulated_follower.add_name_index(follower_number)
                        
                        simulated_process.follower = simulated_follower
                        
                    host_instructions_list.append(simulated_process)
            
            return host_instructions_list
            
        def set_predefined_variables(host, predefined_values, populator, reducer):
            """
            Populate predefined values with expressions 
            and save them as variables in host
            """
            for predefined_value in predefined_values:
                host.set_variable(predefined_value.variable_name, 
                                  populator.populate(predefined_value.expression, 
                                                     host.get_variables(), 
                                                     reducer))
        
        def set_scheduler(host, algorithm):
            """
            Build and set host's scheduler
            """
            host.set_scheduler(scheduler.create(host, algorithm))
        
        built_hosts = []
            
        # Hosts numbers dict keeps the last used number of repeated channel for hosts.
        # For example:
        # run A(*){5} { ... } - will create channels ch1.1, ch1.2, ch1.3, ch1.4 and ch1.5
        # run A(ch1,ch2){2} { ... } - will create channels for next numers which are ch1.6, ch2.6 and ch1.7, ch2.7
        hosts_numbers = {} 
        
        for run_host in version.run_hosts:
            
            if run_host.host_name not in hosts_numbers:
                hosts_numbers[run_host.host_name] = 0
            
            # Create prototype parsed host for this "run host"
            parsed_host = store.find_host(run_host.host_name)
            
            for i in range(0, run_host.repetitions):

                # Build next instructions list for next repeated host                
                instructions_list = build_host_instructions_list(parsed_host, run_host, i)
                
                simulation_host = Host(parsed_host.name, instructions_list)
                    
                # Set the number of host    
                host_number = hosts_numbers[run_host.host_name] + i
                simulation_host.add_name_index(host_number)
                
                # Set scheduler
                set_scheduler(simulation_host, parsed_host.schedule_algorithm)
                
                # Save predefined values as variables
                set_predefined_variables(simulation_host, 
                                         parsed_host.predefined_values, 
                                         populator, 
                                         reducer)
                built_hosts.append(simulation_host)
                
            hosts_numbers[run_host.host_name] += run_host.repetitions 
            
        return built_hosts
    
    def _build_expression_populator(self):
        """
        Build and return object that populates expressions
        with variables' values.
        """
        return expression.Populator()
    
    def _build_expression_checker(self):
        """
        Build and return object that checks the logic value of expressions.
        """
        return expression.Checker()
    
    def _build_expression_reducer(self, equations):
        """
        Build and return object that reduces expressions.
        """
        return expression.Reducer(equations)
    
    def _build_channels(self, store, version, built_hosts):
        """
        Validate, build and return simulation channels build from parsed channels.
        Includes channel repetitions.
        """
        
        def find_channel(channels, name):
            """
            Find channel by name.
            """
            for ch in channels:
                if ch.name == name:
                    return ch
            return None
        
        def find_built_host(built_hosts, name):
            for h in built_hosts:
                if h.name == name:
                    return h
            return None
            
        def find_parsed_process(instructions_list, process_name):
            for instr in instructions_list:
                if not isinstance(instr, HostProcess):
                    continue
                if instr.name == process_name:
                    return instr
            return None
        
        def get_or_create_channel(existing_channels, channel_name):
            """
            Get existing channel by name or create new channel.
            """
            # Find repeated channel with indexes by name 
            channel = find_channel(existing_channels, channel_name)
            # If repeated channel with indexes does not exist
            if channel is None:
                # Find first channel (zero indexes) with original name of repeated channel
                channel = find_channel(existing_channels, original_name(channel_name) + '.0.0')
                channel = channel.clone()
                indexes = name_indexes(channel_name)
                
                for i in range(0, len(indexes)):
                    channel.add_name_index(indexes[i])
                indexes = name_indexes(channel.name)
                
                for i in range(len(indexes), 2):
                    channel.add_name_index(0)
                    
                existing_channels.append(channel)
            return channel
        
        built_channels = []
        original_channels = []
        
        for parsed_channel in store.channels:
            channel = communication.Channel(parsed_channel.name, parsed_channel.buffor_size)
            channel.add_name_index(0)
            channel.add_name_index(0)
            built_channels.append(channel)
            original_channels.append(channel)
            
        # Hosts numbers dict keeps the last used number of repeated channel for hosts.
        # For example:
        # run A(*){5} { ... } - will create channels ch1.1, ch1.2, ch1.3, ch1.4 and ch1.5
        # run A(ch1,ch2){2} { ... } - will create channels for next numers which are ch1.6, ch2.6 and ch1.7, ch2.7
        hosts_numbers = {} 
        
        for run_host in version.run_hosts:
            parsed_host = store.find_host(run_host.host_name)
            # Load all channels names that host can use
            # Firstly it is checked in version, secondly in parsed host
            # List channel_names contains clean channel names (ch1, ch2) 
            channel_names = []
            if run_host.all_channels_active:
                if parsed_host.all_channels_active:
                    channel_names += [ c.original_name() for c in original_channels ]
                else:
                    channel_names += [ c for c in parsed_host.active_channels ]
            else:
                channel_names += [ c for c in run_host.active_channels ]
            # channel_names - clean names of channels that host can use

            if parsed_host.name not in hosts_numbers:
                hosts_numbers[parsed_host.name] = 0
                
            host_first_number = hosts_numbers[parsed_host.name]
            
            # HOST-CHANNELS ASSIGNATION
            # Algorithm:
            # - For each channel name
            #     - Check if channel is repeated
            #     - Get the name of repeated shannel
            #     - If yes:
            #         - Find channel with repeated name with indexes
            #         - If channel with indexes does not exist: 
            #             - Find channel with original name (zero indexes)
            #             - Clone it and add indexes 
            #             - Add new channel to built channels  
            #     - If no 
            
            for channel_name in channel_names:
                
                # Calculate the name of repeated channel (if it is repeated) 
                # with indexes
                repeated_name = ""
                for repeated_channel in run_host.repeated_channels:
                    repeated_channel_basename = original_name(repeated_channel)
                    if channel_name == repeated_channel_basename:
                        indexes = name_indexes(repeated_channel)
                        if len(indexes) > 0:
                            repeated_name = repeated_channel_basename + '.' + str(indexes[0]) + '.0'
                        else:
                            repeated_name += repeated_channel_basename + '.0.0'
                        break
                    
                # If channel is repeated
                if repeated_name != "":
                    
                    channel = get_or_create_channel(built_channels, repeated_name)
                    
                    # Assign channel to hosts created on the base of this "run host"
                    for i in range(0, run_host.repetitions):
                        host_number = host_first_number + i
                        host_name = run_host.host_name + '.' + str(host_number)
                        built_host = find_built_host(built_hosts, host_name)
                        built_host.connect_with_channel(channel)
                    
                else: # Channel is not repeated
                    
                    # New channel is created for each channel name
                    # starting from the lately used number (hosts_numbers)
                    
                    for i in range(0, run_host.repetitions):
                        host_number = host_first_number + i
                        host_name = run_host.host_name + '.' + str(host_number)
                        
                        new_channel_name = channel_name + '.' + str(host_number) + '.0'
                        channel = get_or_create_channel(built_channels, new_channel_name)
                        
                        built_host = find_built_host(built_hosts, host_name)
                        built_host.connect_with_channel(channel)
                        
            # PROCESS-CHANNEL ASSIGNATION
            # Update process-channel assignations in each build host 
            # created on the base of this "run host"
            
            for i in range(0, run_host.repetitions):
                # Process numbers dict keeps the last used number of repeated channel for processes.
                # Works the same like hosts_numbers for hosts.
                processes_numbers = {} 
                
                host_number = host_first_number + i
                host_name = run_host.host_name + '.' + str(host_number)
                
                built_host = find_built_host(built_hosts, host_name)
                parsed_host = store.find_host(built_host.original_name()) 
                
                instruction_index = 0
                for run_process in run_host.run_processes:
                    
                    # Find next process om host's instructions list
                    while instruction_index < len(built_host.instructions_list):
                        if isinstance(built_host.instructions_list[instruction_index], Process):
                            break
                        instruction_index += 1
                        
                    # If no process found raise exception
                    if instruction_index >= len(built_host.instructions_list):
                        raise EnvironmentDefinitionException("Process '%s' not found in host '%s'." % (run_process.process_name, built_host.name))
        
                    if run_process.process_name not in processes_numbers:
                        processes_numbers[run_process.process_name] = 0
                        
                    # Get number of next channel for this process name 
                    process_first_number = processes_numbers[run_process.process_name]
                    
                    # Get process for this "run process"
                    parsed_process = find_parsed_process(parsed_host.instructions_list, run_process.process_name)
                    
                    process_channel_names = []
                    # Get channel names uded by process
                    if parsed_process.all_channels_active:
                        process_channel_names = channel_names
                    else:
                        for process_channel_name in parsed_process.active_channels:
                            process_channel_names.append(process_channel_name)
                            
                    for process_channel_name in process_channel_names:
                        
                        # Calculate the name of repeated channel (if it is repeated) 
                        # with indexes
                        process_channel_repeated_name = "" # Name of repeated channel (may be with indexes)
                        for repeated_channel in run_process.repeated_channels:
                            process_channel_repeated_name_basename = original_name(repeated_channel)
                            if process_channel_name == process_channel_repeated_name_basename:
                                process_channel_repeated_name = repeated_channel
                        
                        # If channel is repeated
                        if process_channel_repeated_name != "":
                            repeated_indexes = name_indexes(process_channel_repeated_name)
                            # If repeated channel is the same, that host is connected with (no indexes)
                            if len(repeated_indexes) == 0:
                                channel = built_host.find_channel(process_channel_repeated_name)
                                # If host does not have channel with this name
                                if  channel is None:
                                    # Generate channel name for this process
                                    ch_name = original_name(process_channel_repeated_name) + '.' +\
                                                str(host_number) + '.' + str(process_first_number)
                                    channel = get_or_create_channel(built_channels, ch_name)
                                    
                            elif len(repeated_indexes) == 1:
                                # Repeated index is the process index in channels (second index)
                                #
                                # Repeated channel has defined numer of channel from channels of current host 
                                # (for different copies of hosts, these are different channels)
                                # For example: Repeated channel name ch1.1 in host A.5 means that all messages
                                # from all copies of process will be sent through channel ch1.x.1 
                                # (where x is the number of channel ch1 in hist A.5 - firstly it would be propably 5)
                                
                                channel = built_host.find_channel(original_name(process_channel_repeated_name))
                                
                                if channel:
                                    # Host has channel with this original name
                                    # Lets check if its second index (process index) is the same 
                                    # with repeated index
                                    
                                    ch_indexes = channel.indexes()
                                    # If channel proces index (2nd) is different from repeated index
                                    if ch_indexes[1] != repeated_indexes[0]:
                                        ch_name = original_name(process_channel_repeated_name) + '.' +\
                                                    str(ch_indexes[0]) + '.' + str(repeated_indexes[0])
                                        channel = get_or_create_channel(built_channels, ch_name)
                                
                                else:
                                    # Channel is not connected with host
                                    # Generate channel name for this process and host
                                    ch_name = original_name(process_channel_repeated_name) + '.' +\
                                                str(host_number) + '.' + str(repeated_indexes[0])
                                    channel = get_or_create_channel(built_channels, ch_name)
                                 
                            elif len(repeated_indexes) == 2:
                                ch_name = process_channel_repeated_name
                                channel = get_or_create_channel(built_channels, ch_name)
                                    
                            else:
                                raise EnvironmentDefinitionException(
                                        'Invalid repeated channel definition %s in process %s (host %s).'\
                                        % (process_channel_repeated_name, run_process.process_name, built_host.name))
                            
                            # Connect repeated channel with all repeated processes
                            # Repeated processes are found as list of next instructions in host's instructions list
                            for i in range(0, run_process.repetitions):
                                built_process = built_host.instructions_list[instruction_index+i]
                                if not isinstance(built_process, Process):
                                    raise EnvironmentDefinitionException(
                                        'Isntruction nr %d of host %s expected to be process.'\
                                        % (instruction_index+i, built_host.name))
                                built_process.connect_with_channel(channel) 
                        
                        else: # Channel is not repeated
                            
                                current_host_channel_index = host_number
                                
                                # Check if host is connected with any channel of this original name
                                channel = built_host.find_channel(original_name(process_channel_name))
                                if channel:
                                    # Get current host channel index 
                                    # from the second index of found channel indexes
                                    ch_indexes = channel.indexes()
                                    current_host_channel_index = ch_indexes[0]
                                    
                                for i in range(0, run_process.repetitions):
                                    # Generate the channel name for all repetitions
                                    process_number = process_first_number + i
                                    
                                    # Search for channel with generated name or create one
                                    ch_name = original_name(process_channel_name) + '.' +\
                                            str(current_host_channel_index) + '.' + str(process_number)
                                    channel = get_or_create_channel(built_channels, ch_name)
                                    
                                    # Connect channel with process  
                                    built_process = built_host.instructions_list[instruction_index+i]
                                    
                                    if not isinstance(built_process, Process):
                                        raise EnvironmentDefinitionException(
                                            'Isntruction nr %d of host %s expected to be process.'\
                                            % (instruction_index+i, built_host.name))
                                    built_process.connect_with_channel(channel)
                    
                    # Update process numbers dict
                    processes_numbers[run_process.process_name] += run_process.repetitions
                    instruction_index += run_process.repetitions
        
            # Update hosts numbers dict
            hosts_numbers[run_host.host_name] += run_host.repetitions
                                
        return built_channels
    
    def _build_channels_manager(self, channels):
        """ 
        Build channels manager
        """
        return communication.Manager(channels)
        
    def _build_metrics_manager(self, store, hosts, version):
        """
        Build and return metrics manager.
        """
        host_metrics = []
        
        for metrics_data in store.metrics_datas:
            
            blocks = []
            # Build list of metrics blocks
            for block in metrics_data.blocks:
                
                params = block.header.params[1:]
                service_params = block.header.services_params
                metrics_block = metrics.Block(params, service_params)
                
                for metric in block.metrics:
                    m = metrics.Metric(metric.arguments[0], metric.arguments[1:len(params)+1], 
                                       metric.arguments[len(params)+1:])
                    metrics_block.add_metric(m)
                blocks.append(metrics_block)
            
            # get or create host metrics with given name
            hm = None
            for existing_hm in host_metrics:
                if existing_hm.name == metrics_data.name:
                    hm = existing_hm
                    break
            if hm is None:
                hm = metrics.HostMetrics(metrics_data.name)
                host_metrics.append(hm)
            
            if metrics_data.plus or metrics_data.star:
                if metrics_data.plus:
                    hm.plus_blocks = blocks
                else: # star
                    hm.star_blocks = blocks
                    
                # If host metrics does not have normal block
                # Search for them and assign if found in host metrics 
                # with simple name (not qualified like hm1.1)
                if len(hm.normal_blocks) == 0:
                    hm_original_name = original_name(hm.name)
                    hm_original = None
                    
                    for existing_hm in host_metrics:
                        if original_name(existing_hm.name) == hm_original_name \
                            and existing_hm != hm:
                            hm_original = existing_hm
                            break
                        
                    if hm_original:
                        hm.normal_blocks = hm_original.normal_blocks
                        
            else: # metrics_data normal
                hm_original_name = original_name(hm.name)
                
                # Assign normal block to all host metrics with the same original name
                for existing_hm in host_metrics:
                    if original_name(existing_hm.name) == hm_original_name:
                        # Assign notmal block to all host metrics with the same original name
                        # (including the current one - possibly created)
                        existing_hm.normal_blocks = blocks
                        
        # Connect host metrics with hosts
        for metrics_set in version.metrics_sets:
            for h in hosts:
                if h.original_name() == metrics_set.host_name:
                    for host_metric in host_metrics:
                        if host_metric.name == metrics_set.configuration_name:
                            host_metric.connected_hosts.append(h)
                            break
        
        return metrics.Manager(host_metrics)
    
    def _build_context(self, store, version):
        """
        Builds context with initial state.
        """
        functions = self._build_functions(store)
        equations = self._build_equations(store, functions)
        expression_reducer = self._build_expression_reducer(equations)
        expression_populator = self._build_expression_populator()
        expression_checker = self._build_expression_checker()
        hosts = self._build_hosts(store, version, functions, 
                                  expression_populator, expression_reducer)
        channels = self._build_channels(store, version, hosts)

        c = state.Context(version)
        c.functions = functions
        c.hosts = hosts
        c.expression_reducer = expression_reducer
        c.expression_checker = expression_checker
        c.expression_populator = expression_populator
        c.metrics_manager = self._build_metrics_manager(store, hosts, version);
        c.channels_manager = self._build_channels_manager(channels)
        return c
    
    def build_executor(self):
        """
        Creates executor for simulation
        """
        e = Executor()
        e.append_instruction_executor(AssignmentInstructionExecutor())
        e.append_instruction_executor(CallFunctionInstructionExecutor())
        e.append_instruction_executor(ProcessInstructionExecutor())
        e.append_instruction_executor(SubprocessInstructionExecutor())
        e.append_instruction_executor(CommunicationInstructionExecutor())
        e.append_instruction_executor(FinishInstructionExecutor())
        e.append_instruction_executor(ContinueInstructionExecutor())
        e.append_instruction_executor(IfInstructionExecutor())
        e.append_instruction_executor(WhileInstructionExecutor())
        return e
    
    def build_simulator(self, store, version):
        """
        Creates simulator for particular version.
        """
        sim = Simulator(self._build_context(store, version))
        sim.set_executor(self.build_executor())
        return sim
    
    def build_model_parser(self, store, modules):
        """
        Builder parser that parses model written in QoPML
        and populates the store.
        """
        from aqopa.model.parser.lex_yacc import LexYaccParser
        from aqopa.model.parser.lex_yacc.grammar import main,\
                functions, channels, equations, expressions, instructions,\
                hosts, metrics
        
        parser = LexYaccParser()
        parser.set_store(store) \
                .add_extension(main.ModelParserExtension()) \
                .add_extension(functions.ModelParserExtension()) \
                .add_extension(channels.ModelParserExtension()) \
                .add_extension(equations.ModelParserExtension()) \
                .add_extension(expressions.ModelParserExtension()) \
                .add_extension(instructions.ModelParserExtension()) \
                .add_extension(hosts.ModelParserExtension()) 
                
        for m in modules:
            parser = m.extend_model_parser(parser)
                
        return parser.build()
    
    def build_metrics_parser(self, store, modules):
        """
        Builder parser that parses metrics written in QoPML
        and populates the store.
        """
        from aqopa.model.parser.lex_yacc import LexYaccParser
        from aqopa.model.parser.lex_yacc.grammar import main, metrics
        
        parser = LexYaccParser()
        parser.set_store(store)\
                .add_extension(main.MetricsParserExtension())\
                .add_extension(metrics.MetricsParserExtension())
                
        for m in modules:
            parser = m.extend_metrics_parser(parser)
                
        return parser.build()
    
    def build_config_parser(self, store, modules):
        """
        Builder parser that parses config written in QoPML
        and populates the store.
        """
        from aqopa.model.parser.lex_yacc import LexYaccParser
        from aqopa.model.parser.lex_yacc.grammar import versions, main
        
        parser = LexYaccParser()
        parser.set_store(store)\
                .add_extension(main.ConfigParserExtension())\
                .add_extension(versions.ConfigParserExtension())
                
        for m in modules:
            parser = m.extend_config_parser(parser)
                
        return parser.build()
    
class Interpreter():
    """
    Interpreter is responsible for parsing the model,
    creating the environment for simulations,
    manipulating selected models.
    """
    
    def __init__(self, builder=None, model_as_text="", 
                 metrics_as_text="", config_as_text=""):
        self.builder = builder if builder is not None else Builder()
        self.model_as_text = model_as_text
        self.metrics_as_text = metrics_as_text 
        self.config_as_text = config_as_text
        
        self.store = None
        
        self._modules = []
        
    def set_qopml_model(self, model_as_text):
        """
        Set qopml model that will be interpreted.
        """
        self.model_as_text = model_as_text
        return self
    
    def set_qopml_metrics(self, metrics_as_text):
        """
        Set qopml metrics that will be interpreted.
        """
        self.metrics_as_text = metrics_as_text
        return self
    
    def set_qopml_config(self, config_as_text):
        """
        Set qopml configuration that will be interpreted.
        """
        self.config_as_text = config_as_text
        return self
    
    def register_qopml_module(self, qopml_module):
        """
        Registers new module
        """
        if qopml_module in self._modules:
            raise EnvironmentDefinitionException(u"QoPML Module '%s' is already registered." % unicode(qopml_module))
        self._modules.append(qopml_module)
        return self
        
    def parse(self):
        """
        Parses the model from model_as_text field and populates the store.
        """
        if len(self.model_as_text) == 0:
            raise EnvironmentDefinitionException("QoPML Model not provided.")
    
        self.store = self.builder.build_store()
    
        parser = self.builder.build_model_parser(self.store, self._modules)
        parser.parse(self.model_as_text)
        if len(parser.get_syntax_errors()) > 0:
            raise ModelParserException('Invalid syntax.', syntax_errors=parser.get_syntax_errors())
    
        parser = self.builder.build_metrics_parser(self.store, self._modules)
        parser.parse(self.metrics_as_text)
        if len(parser.get_syntax_errors()) > 0:
            raise MetricsParserException('Invalid syntax.', syntax_errors=parser.get_syntax_errors())
    
        parser = self.builder.build_config_parser(self.store, self._modules)
        parser.parse(self.config_as_text)
        if len(parser.get_syntax_errors()) > 0:
            raise ConfigurationParserException('Invalid syntax.', syntax_errors=parser.get_syntax_errors())
    
    def install_modules(self, simulator):
        """ """
        raise NotImplementedError()
    
    def run(self):
        """ Runs all simulations """
        raise NotImplementedError()
            
class ConsoleInterpreter(Interpreter):
    
    def __init__(self, builder=None, model_as_text="", 
                 metrics_as_text="", config_as_text=""):
        Interpreter.__init__(self, builder, model_as_text, metrics_as_text, config_as_text)
        
        self.simulators = []

    def save_states_to_file(self, simulator):
        """ 
        Tells simulator to save states flow to file.
        """
        f = open('VERSION_%s_STATES_FLOW' % simulator.context.version.name, 'w')
        simulator.get_executor().prepend_instruction_executor(PrintExecutor(f))      

    def prepare(self):
        """ Prepares for run """
        
        for version in self.store.versions:
            simulator = self.builder.build_simulator(self.store, version)
            self.simulators.append(simulator)
            self.install_modules(simulator)
    
    def install_modules(self, simulator):
        """ """
        for m in self._modules:
            m.install_console(simulator)
        
    def run(self):
        """ Runs all simulations """

        for s in self.simulators:
            s.prepare()
            s.run()
            self.on_finished(s)
            
    def on_finished(self, simulator):
        pass


class GuiInterpreter(Interpreter):
    
    def __init__(self, builder=None, model_as_text="", 
                 metrics_as_text="", config_as_text=""):
        Interpreter.__init__(self, builder=builder, 
                             model_as_text=model_as_text, 
                             metrics_as_text=metrics_as_text, 
                             config_as_text=config_as_text)
        self.simulators = []

    def prepare(self):
        """ Prepares for run """
        for version in self.store.versions:
            simulator = self.builder.build_simulator(self.store, version)
            self.install_modules(simulator)
            self.simulators.append(simulator)
    
    def install_modules(self, simulator):
        """ """
        for m in self._modules:
            m.install_gui(simulator)
            
    def run_simulation(self, simulator):
        """ """
        simulator.prepare()
        simulator.run()
        return simulator
